from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraServiceSite(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_services_sites'
    id_service_site = Column(Integer, Sequence('id_service_site_sequence'), primary_key=True, autoincrement=True)
    id_service = Column(Integer, ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_site = Column(Integer, ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=False)

    service_site_service = relationship("TeraService", viewonly=True)
    service_site_site = relationship("TeraSite", viewonly=True)

    def __init__(self):
        pass

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_site_service', 'service_site_site'])

        if minimal:
            ignore_fields.extend([])

        return super().to_json(ignore_fields=ignore_fields)

    @staticmethod
    def get_services_for_site(id_site: int, with_deleted: bool = False):
        return TeraServiceSite.query.execution_options(include_deleted=with_deleted).filter_by(id_site=id_site).all()

    @staticmethod
    def get_sites_for_service(id_service: int, with_deleted: bool = False):
        return TeraServiceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=id_service).all()

    @staticmethod
    def get_service_site_by_id(service_site_id: int, with_deleted: bool = False):
        return TeraServiceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_site=service_site_id).first()

    @staticmethod
    def get_service_site_for_service_site(site_id: int, service_id: int, with_deleted: bool = False):
        return TeraServiceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id, id_service=service_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraSite import TeraSite

            site1 = TeraSite.get_site_by_sitename('Default Site')
            site2 = TeraSite.get_site_by_sitename('Top Secret Site')

            service_rehab = TeraService.get_service_by_key('VideoRehabService')
            service_filetransfer = TeraService.get_service_by_key('FileTransferService')
            service_logging = TeraService.get_service_by_key('LoggingService')

            service_site = TeraServiceSite()
            service_site.id_site = site1.id_site
            service_site.id_service = service_rehab.id_service
            TeraServiceSite.db().session.add(service_site)

            service_site = TeraServiceSite()
            service_site.id_site = site1.id_site
            service_site.id_service = service_filetransfer.id_service
            TeraServiceSite.db().session.add(service_site)

            service_site = TeraServiceSite()
            service_site.id_site = site1.id_site
            service_site.id_service = service_logging.id_service
            TeraServiceSite.db().session.add(service_site)

            service_site = TeraServiceSite()
            service_site.id_site = site2.id_site
            service_site.id_service = service_filetransfer.id_service
            TeraServiceSite.db().session.add(service_site)

            TeraServiceSite.db().session.commit()
        else:
            # Automatically associate services that are in a project to that site
            from opentera.db.models.TeraServiceProject import TeraServiceProject
            for sp in TeraServiceProject.query_with_filters():
                project_site_id = sp.service_project_project.id_site
                if not TeraServiceSite.get_service_site_for_service_site(site_id=project_site_id,
                                                                         service_id=
                                                                         sp.service_project_service.id_service):
                    # No association - create a new one
                    service_site = TeraServiceSite()
                    service_site.id_site = project_site_id
                    service_site.id_service = sp.service_project_service.id_service
                    TeraServiceSite.db().session.add(service_site)
                    TeraServiceSite.db().session.commit()

    @staticmethod
    def delete_with_ids(service_id: int, site_id: int, autocommit: bool = True):
        delete_obj = TeraServiceSite.query.filter_by(id_service=service_id, id_site=site_id).first()
        if delete_obj:
            TeraServiceSite.delete(delete_obj.id_service_site, autocommit=autocommit)

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True):
        from opentera.db.models.TeraServiceProject import TeraServiceProject
        # Delete all association with projects for that site
        delete_obj = TeraServiceSite.query.filter_by(id_service_site=id_todel).first()

        if delete_obj:
            projects = TeraServiceProject.get_projects_for_service(delete_obj.id_service)
            for service_project in projects:
                if service_project.service_project_project.id_site == delete_obj.id_site:
                    TeraServiceProject.delete(service_project.id_service_project, autocommit=autocommit)

            from opentera.db.models.TeraSessionTypeSite import TeraSessionTypeSite
            session_types = TeraSessionTypeSite.get_session_type_site_for_site_and_service(
                site_id=delete_obj.id_site, service_id=delete_obj.id_service)
            for session_type in session_types:
                TeraSessionTypeSite.delete(session_type.id_session_type_site, autocommit=autocommit)

            # Ok, delete it
            super().delete(id_todel, autocommit=autocommit)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        from opentera.db.models.TeraServiceProject import TeraServiceProject
        projects = TeraServiceProject.get_projects_for_service(self.id_service, with_deleted=with_deleted)
        for service_project in projects:
            if service_project.service_project_project.id_site == self.id_site:
                if service_project.delete_check_integrity(with_deleted=with_deleted):
                    return IntegrityError('Still have sessions with that service', self.id_service,
                                          't_sessions')
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.config.ConfigManager import ConfigManager
from opentera.logging.LoggingClient import LoggingClient
import redis
import jwt
from typing import Set, Any


class DisabledTokenStorage:
    """
    This class is used to store disabled tokens in redis.
    """
    def __init__(self, redis_key: str):
        self.redis_key = redis_key
        self.redis_client = None
        self.token_key = None
        self.logging_client = None

    def config(self, config: ConfigManager, token_key: str) -> None:
        self.token_key = token_key
        self.redis_client = redis.Redis(host=config.redis_config['hostname'],
                                        port=config.redis_config['port'],
                                        username=config.redis_config['username'],
                                        password=config.redis_config['password'],
                                        db=config.redis_config['db'])

        self.logging_client = LoggingClient(config=config.redis_config,
                                            client_name="DisabledTokenStorage-" + str(self.redis_key))

        # Remove all expired tokens
        self.remove_all_expired_tokens(self.token_key)

    def add_disabled_token(self, token) -> Any:
        # Add token to set
        return self.redis_client.sadd(self.redis_key, token)

    def get_disabled_tokens(self) -> Set[str]:
        # Get all elements from the set
        return self.redis_client.smembers(self.redis_key)

    def is_disabled_token(self, token) -> bool:
        # Check if token is in set
        if not token:
            return True
        return self.redis_client.sismember(self.redis_key, token)

    def clear_all_disabled_tokens(self) -> Any:
        # Clear set
        return self.redis_client.delete(self.redis_key)

    def remove_disabled_token(self, token) -> Any:
        # Remove token from set
        self.redis_client.srem(self.redis_key, token)

    def remove_all_expired_tokens(self, key) -> Set[str]:
        to_be_removed: set = set()
        for token in self.redis_client.smembers(self.redis_key):

            if token is None:
                continue

            if len(token) == 0:
                self.redis_client.srem(self.redis_key, token)
                continue

            try:
                token_dict = jwt.decode(token, key, algorithms='HS256')
                # Expired tokens will throw exception.
                # If we continue here, tokens have a valid expiration time.
                # We should stop looking for expired tokens since they are added chronologically
                break
            except jwt.exceptions.ExpiredSignatureError as e:
                # Remove expired token
                to_be_removed.add(token)
                self.redis_client.srem(self.redis_key, token)
            except jwt.exceptions.InvalidSignatureError as e:
                # Token was signed with a different key, remove it
                self.redis_client.srem(self.redis_key, token)
                self.logging_client.log_error('DisabledTokenStorage.remove_all_expired_tokens', str(e))
            except jwt.exceptions.PyJWTError as e:
                self.logging_client.log_error('DisabledTokenStorage.remove_all_expired_tokens', str(e))
                continue

        # Return removed tokens
        return to_be_removed



from opentera.db.models.TeraServerSettings import TeraServerSettings
from opentera import OpenTeraServerVersion
import json


class ClientVersions:
    def __init__(self, **kwargs):
        self.name = kwargs.get('client_name', None)
        self.description = kwargs.get('client_description', None)
        self.version = kwargs.get('client_version', None)
        self.documentation_url = kwargs.get('client_documentation_url', None)
        self.windows_download_url = kwargs.get('client_windows_download_url', None)
        self.mac_download_url = kwargs.get('client_mac_download_url', None)
        self.linux_download_url = kwargs.get('client_linux_download_url', None)

    @property
    def client_name(self):
        return self.name

    @client_name.setter
    def client_name(self, name: str):
        self.name = name

    @property
    def client_description(self):
        return self.description

    @client_description.setter
    def client_description(self, description: str):
        self.description = description

    @property
    def client_version(self):
        return self.version

    @client_version.setter
    def client_version(self, version: str):
        self.version = version

    @property
    def client_documentation_url(self):
        return self.documentation_url

    @client_documentation_url.setter
    def client_documentation_url(self, value: str):
        self.documentation_url = value

    @property
    def client_windows_download_url(self):
        return self.windows_download_url

    @client_windows_download_url.setter
    def client_windows_download_url(self, value: str):
        self.windows_download_url = value

    @property
    def client_mac_download_url(self):
        return self.mac_download_url

    @client_mac_download_url.setter
    def client_mac_download_url(self, value: str):
        self.mac_download_url = value

    @property
    def client_linux_download_url(self):
        return self.linux_download_url

    @client_linux_download_url.setter
    def client_linux_download_url(self, value: str):
        self.linux_download_url = value

    def from_dict(self, value: dict):
        if 'client_name' in value:
            self.client_name = value['client_name']
        if 'client_description' in value:
            self.client_description = value['client_description']
        if 'client_version' in value:
            self.client_version = value['client_version']
        if 'client_documentation_url' in value:
            self.client_documentation_url = value['client_documentation_url']
        if 'client_windows_download_url' in value:
            self.client_windows_download_url = value['client_windows_download_url']
        if 'client_mac_download_url' in value:
            self.client_mac_download_url = value['client_mac_download_url']
        if 'client_linux_download_url' in value:
            self.client_linux_download_url = value['client_linux_download_url']

    def to_dict(self):
        return {'client_name': self.client_name,
                'client_description': self.client_description,
                'client_version': self.client_version,
                'client_documentation_url': self.client_documentation_url,
                'client_windows_download_url': self.client_windows_download_url,
                'client_mac_download_url': self.client_mac_download_url,
                'client_linux_download_url': self.client_linux_download_url}

    @classmethod
    def get_json_schema(cls):
        model_name = cls.__name__
        # Properties dict
        pr_dict = dict()
        # Fill properties
        pr_dict['client_name'] = {'type': 'string', 'required': True}
        pr_dict['client_description'] = {'type': 'string', 'required': True}
        pr_dict['client_version'] = {'type': 'string', 'required': True}
        pr_dict['client_documentation_url'] = {'type': 'string', 'required': True}
        pr_dict['client_windows_download_url'] = {'type': 'string', 'required': False}
        pr_dict['client_mac_download_url'] = {'type': 'string', 'required': False}
        pr_dict['client_linux_download_url'] = {'type': 'string', 'required': False}

        schema = {model_name: {'properties': pr_dict, 'type': 'object'}}
        return schema

    def __repr__(self):
        return '<ClientVersions ' + json.dumps(self.to_dict()) + ' >'


class TeraVersions:
    def __init__(self):
        self.server_version = str(OpenTeraServerVersion.opentera_server_version_string)
        self.server_major_version = OpenTeraServerVersion.opentera_server_major_version
        self.server_minor_version = OpenTeraServerVersion.opentera_server_minor_version
        self.server_patch_version = OpenTeraServerVersion.opentera_server_patch_version
        self.clients = dict()
        # Hard coding OpenTeraPlus for now
        # Will be overwritten if necessary
        self.clients['OpenTeraPlus'] = ClientVersions(client_name='OpenTeraPlus',
                                                      client_description='OpenTeraPlus Qt Client',
                                                      client_version='1.0.0',
                                                      client_documentation_url=
                                                      'https://github.com/introlab/openteraplus')

    def get_client_version_with_name(self, name: str):
        if name in self.clients:
            return self.clients[name]
        return None

    def set_client_version_with_name(self, name: str, client_version: ClientVersions):
        self.clients[name] = client_version

    @property
    def version_string(self):
        return self.server_version

    @property
    def major_version(self):
        return self.server_major_version

    @property
    def minor_version(self):
        return self.server_minor_version

    @property
    def patch_version(self):
        return self.server_patch_version

    @property
    def client_dict(self):
        return self.clients

    def to_dict(self) -> dict:
        output = {'version_string': self.server_version,
                  'version_major': self.server_major_version,
                  'version_minor': self.server_minor_version,
                  'version_patch': self.server_patch_version,
                  'clients': {}}
        # Output clients dict
        for client_name in self.clients:
            output['clients'][client_name] = self.clients[client_name].to_dict()
        return output

    def from_dict(self, value: dict):
        # We do not want to load version from DB
        # TODO can we do better ?
        # if 'version_string' in value:
        #     self.server_version = value['version_string']
        # if 'version_major' in value:
        #     self.server_major_version = value['version_major']
        # if 'version_minor' in value:
        #     self.server_minor_version = value['version_minor']
        # if 'version_patch' in value:
        #     self.server_patch_version = value['version_patch']
        if 'clients' in value:
            # We have a list of clients dict in value['clients']
            # Make sure we have a dict and not a list (in old implementation)
            if isinstance(value['clients'], dict):
                for client_name in value['clients']:
                    client_version = ClientVersions()
                    client_version.from_dict(value['clients'][client_name])
                    # Put in the dict, overwriting if client have the same name
                    # TODO can we do better?
                    self.clients[client_version.client_name] = client_version

    def save_to_db(self):
        TeraServerSettings.set_server_setting(TeraServerSettings.ServerVersions, json.dumps(self.to_dict()))

    def load_from_db(self):
        settings = TeraServerSettings.get_server_setting_value(TeraServerSettings.ServerVersions)
        if settings:
            self.from_dict(json.loads(settings))
            return True
        return False

    def __repr__(self):
        return '<TeraVersions: ' + json.dumps(self.to_dict()) + ' >'


# if __name__ == '__main__':
#     versions = TeraVersions()
#     versions_dict = versions.to_dict()
#     versions2 = TeraVersions()
#     versions2.from_dict(versions_dict)
#     print(versions.get_client_version_with_name('OpenTeraPlus'))
#     print(versions.get_client_version_with_name('Unknown'))
#     print(versions)
#     print(versions2)
#     print(ClientVersions.get_json_schema())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: LeaveSessionEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17LeaveSessionEvent.proto\x12\x11opentera.protobuf\"\x8d\x01\n\x11LeaveSessionEvent\x12\x14\n\x0csession_uuid\x18\x01 \x01(\t\x12\x14\n\x0cservice_uuid\x18\x02 \x01(\t\x12\x1c\n\x14leaving_participants\x18\x03 \x03(\t\x12\x15\n\rleaving_users\x18\x04 \x03(\t\x12\x17\n\x0fleaving_devices\x18\x05 \x03(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'LeaveSessionEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LEAVESESSIONEVENT._serialized_start=47
  _LEAVESESSIONEVENT._serialized_end=188
# @@protoc_insertion_point(module_scope)



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from datetime import datetime
import json
import jsonschema


class TeraServiceConfig(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_services_configs'
    id_service_config = Column(Integer, Sequence('id_service_config_sequence'), primary_key=True,
                               autoincrement=True)
    id_service = Column(Integer, ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_user = Column(Integer, ForeignKey('t_users.id_user', ondelete='cascade'), nullable=True)
    id_participant = Column(Integer, ForeignKey('t_participants.id_participant', ondelete='cascade'), nullable=True)
    id_device = Column(Integer, ForeignKey('t_devices.id_device', ondelete='cascade'), nullable=True)
    service_config_config = Column(String, nullable=False, default='{}')

    service_config_service = relationship("TeraService", viewonly=True)
    service_config_user = relationship("TeraUser", viewonly=True)
    service_config_participant = relationship('TeraParticipant', viewonly=True)
    service_config_device = relationship('TeraDevice', viewonly=True)
    service_config_specifics = relationship('TeraServiceConfigSpecific', cascade='delete')

    def to_json(self, ignore_fields=None, minimal=False, specific_id=None):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_config_service', 'service_config_user', 'service_config_participant',
                              'service_config_device', 'service_config_specifics'])

        if minimal:
            ignore_fields.extend([])

        json_config = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            json_config['service_config_name'] = self.service_config_service.service_name
            if self.version_id:  # No version ID can occur on temporary objects
                json_config['service_config_last_update_time'] = self.get_last_update_datetime().isoformat()
            else:
                json_config['service_config_last_update_time'] = None

        # Filter null ids
        if not self.id_user:
            del json_config['id_user']

        if not self.id_participant:
            del json_config['id_participant']

        if not self.id_device:
            del json_config['id_device']

        # Get correct config
        if specific_id and self.has_specific_config_for_specific_id(specific_id=specific_id):
            json_config['service_config_config'] = self.get_config_for_specific_id(specific_id=specific_id)
            json_config['service_config_specific_id'] = specific_id
        else:
            json_config['service_config_config'] = self.get_global_config()
            json_config['service_config_specific_id'] = None

        return json_config

    def from_json(self, json_values: dict):
        super().from_json(json=json_values, ignore_fields=['service_config_config', 'id_specific'])

        if 'service_config_config' in json_values:
            if 'id_specific' not in json_values:
                # Specifc config can't be handled here, since id_service_config might be equal to 0...
                #     self.set_config_for_specific_id(json_values['id_specific'], json_values['service_config_config'])
                # else:
                self.set_global_config(json_values['service_config_config'])

    @staticmethod
    def get_service_config_by_id(s_id: int, with_deleted: bool = False):
        return TeraServiceConfig.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_config=s_id).first()

    @staticmethod
    def get_service_config_for_service(service_id: int, with_deleted: bool = False):
        return TeraServiceConfig.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id).all()

    @staticmethod
    def get_service_config_for_service_for_user(service_id: int, user_id: int, with_deleted: bool = False):
        return TeraServiceConfig.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, id_user=user_id).first()

    @staticmethod
    def get_service_config_for_service_for_participant(service_id: int, participant_id: int,
                                                       with_deleted: bool = False):
        return TeraServiceConfig.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, id_participant=participant_id).first()

    @staticmethod
    def get_service_config_for_service_for_device(service_id: int, device_id: int, with_deleted: bool = False):
        return TeraServiceConfig.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, id_device=device_id).first()

    def has_specific_config_for_specific_id(self, specific_id: int) -> bool:
        for specific in self.service_config_specifics:
            if specific.service_config_specific_id == specific_id:
                return True
        return False

    def get_last_update_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.version_id/1000)

    def config_is_valid(self) -> bool:
        if self.service_config_config is None:
            return True  # Will use default value

        # Check if the config is a correct json format
        if isinstance(self.service_config_config, str):
            try:
                json.loads(self.service_config_config)
            except ValueError as err:
                raise err

        return True

    def get_global_config(self) -> dict:
        if self.service_config_config is None:
            return dict()

        try:
            config_json = json.loads(self.service_config_config)
        except ValueError as err:
            raise err
        return config_json

    def get_config_for_specific_id(self, specific_id: str) -> dict:
        if self.service_config_config is None:
            return dict()

        try:
            config_json = json.loads(self.service_config_config)
        except ValueError as err:
            raise err

        # No specific config at all - return global config
        if self.service_config_specifics is None or len(self.service_config_specifics) == 0:
            return config_json

        # Check if we have a specific config
        for specific in self.service_config_specifics:
            if specific.service_config_specific_id == specific_id:
                # Complete or overwrite values in current config
                try:
                    specific_config = json.loads(specific.service_config_specific_config)
                except ValueError as err:
                    raise err
                for specific_key, specific_value in specific_config.items():
                    config_json[specific_key] = specific_value

        return config_json

    def set_config_for_specific_id(self, specific_id: str, config: str):
        if isinstance(config, dict):
            try:
                config = json.dumps(config)
            except ValueError as err:
                raise err

        #  Set or update specific config
        from opentera.db.models.TeraServiceConfigSpecific import TeraServiceConfigSpecific
        TeraServiceConfigSpecific.insert_or_update_specific_config(service_config_id=self.id_service_config,
                                                                   specific_id=specific_id, config=config)

    def set_global_config(self, config: str):
        if isinstance(config, dict):
            try:
                config = json.dumps(config)
            except ValueError as err:
                raise err

        self.service_config_config = config

    @staticmethod
    def get_empty_config() -> dict:
        # return {'Globals': {}, 'Specifics': {}}
        return dict()

    def get_specific_ids_list(self) -> list:
        return [specific_config.service_config_specific_id for specific_config in self.service_config_specifics]

    @staticmethod
    def create_defaults(test=False):
        if test:
            from .TeraUser import TeraUser
            from .TeraDevice import TeraDevice
            from .TeraService import TeraService
            from .TeraParticipant import TeraParticipant
            from .TeraServiceConfigSpecific import TeraServiceConfigSpecific

            new_config = TeraServiceConfig()
            new_config.id_user = TeraUser.get_user_by_id(1).id_user
            new_config.id_service = TeraService.get_openteraserver_service().id_service
            new_config.service_config_config = '{"notification_sounds": true}'
            TeraServiceConfig.db().session.add(new_config)

            new_config = TeraServiceConfig()
            new_config.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
            new_config.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            new_config.service_config_config = '{"default_muted": false, "view_local": true}'
            TeraServiceConfig.db().session.add(new_config)
            TeraServiceConfig.db().session.commit()

            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = new_config.id_service_config
            new_specific_config.service_config_specific_id = 'pc-001'
            new_specific_config.service_config_specific_config = '{"default_muted": true}'
            TeraServiceConfig.db().session.add(new_specific_config)

            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = new_config.id_service_config
            new_specific_config.service_config_specific_id = 'pc-002'
            new_specific_config.service_config_specific_config = '{"default_muted": true, "view_local": false}'
            TeraServiceConfig.db().session.add(new_specific_config)

            new_config = TeraServiceConfig()
            new_config.id_device = TeraDevice.get_device_by_name('Apple Watch #W05P1').id_device
            new_config.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            new_config.service_config_config = '{"delete_from_device": true}'
            TeraServiceConfig.db().session.add(new_config)

            TeraServiceConfig.db().session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Validate that the config we have is valid
        if 'service_config_config' in values:
            if 'id_specific' in values:
                # We have a specific config...
                config = TeraServiceConfig.get_service_config_by_id(values['id_service_config'])
                if not config:
                    return False

                config.set_config_for_specific_id(specific_id=values['id_specific'],
                                                  config=values['service_config_config'])

                del values['id_specific']
            else:
                # We have a global config...
                config = TeraServiceConfig.get_service_config_by_id(values['id_service_config'])
                if not config:
                    return False

                config.set_global_config(config=values['service_config_config'])

            if not config.config_is_valid():
                return False

            # Already updated - remove from list
            del values['service_config_config']

        super().update(update_id=update_id, values=values)
        return True

    @classmethod
    def insert(cls, db_object):
        # Validate that the config we have is valid
        try:
            config_valid = db_object.config_is_valid()
        except (ValueError, jsonschema.exceptions.ValidationError) as err:
            raise err

        if not config_valid:
            return False

        super().insert(db_object)
        return True




"""
Main query rewriter logic.
Adapted from:
https://github.com/flipbit03/sqlalchemy-easy-softdelete
"""

from typing import TypeVar, Union

from sqlalchemy import Table
from sqlalchemy.orm import FromStatement
from sqlalchemy.orm.util import _ORMJoin
from sqlalchemy.sql import Alias, CompoundSelect, Join, Select, Subquery, TableClause
from sqlalchemy.sql.elements import TextClause

Statement = TypeVar('Statement', bound=Union[Select, FromStatement])


class SoftDeleteQueryRewriter:
    """Rewrites SQL statements based on configuration."""

    def __init__(self, deleted_field_name: str, disable_soft_delete_option_name: str):
        """
        Instantiate a new query rewriter.

        Params:

        deleted_field_name:
            The name of the field that should be present in a table for soft-deletion
            rewriting to occur

        disable_soft_delete_option_name:
            Execution option name (to use with .execution_options(xxxx=True) to disable
            soft deletion rewriting in a query

        """
        self.deleted_field_name = deleted_field_name
        self.disable_soft_delete_option_name = disable_soft_delete_option_name

    def rewrite_statement(self, stmt: Statement) -> Statement:
        """Rewrite a single SQL-like Statement."""
        if isinstance(stmt, Select):
            return self.rewrite_select(stmt)

        if isinstance(stmt, FromStatement):
            # Explicitly protect against INSERT with RETURNING
            if not isinstance(stmt.element, Select):
                return stmt
            stmt.element = self.rewrite_select(stmt.element)
            return stmt

        raise NotImplementedError(f"Unsupported statement type \"{(type(stmt))}\"!")

    def rewrite_select(self, stmt: Select) -> Select:
        """Rewrite a Select Statement."""
        # if the user tagged this query with an execution_option to disable soft-delete filtering
        # simply return back the same stmt
        if stmt.get_execution_options().get(self.disable_soft_delete_option_name):
            return stmt

        for from_obj in stmt.get_final_froms():
            stmt = self.analyze_from(stmt, from_obj)

        return stmt

    def rewrite_compound_select(self, stmt: CompoundSelect) -> CompoundSelect:
        """Rewrite a Compound Select Statement."""
        # This needs to be done by array slice referencing instead of
        # a direct reassignment because the reassignment would not substitute the
        # value which is inside the CompoundSelect "by reference"
        for i in range(len(stmt.selects)):
            stmt.selects[i] = self.rewrite_select(stmt.selects[i])
        return stmt

    def rewrite_element(self, subquery: Subquery) -> Subquery:
        """Rewrite an object with a `.element` attribute and patch the query inside it."""
        if isinstance(subquery.element, CompoundSelect):
            subquery.element = self.rewrite_compound_select(subquery.element)
            return subquery

        if isinstance(subquery.element, Select):
            subquery.element = self.rewrite_select(subquery.element)
            return subquery

        raise NotImplementedError(f"Unsupported object \"{(type(subquery.element))}\" in subquery.element")

    def rewrite_from_orm_join(self, stmt: Select, join_obj: Union[_ORMJoin, Join]) -> Select:
        # Recursive cases (multiple joins)
        if isinstance(join_obj.left, _ORMJoin) or isinstance(join_obj.left, Join):
            stmt = self.rewrite_from_orm_join(stmt, join_obj.left)

        if isinstance(join_obj.right, _ORMJoin) or isinstance(join_obj.right, Join):
            stmt = self.rewrite_from_orm_join(stmt, join_obj.right)

        if isinstance(join_obj.left, Subquery):
            for from_obj in join_obj.left.element.get_final_froms():
                stmt = self.analyze_from(stmt, from_obj)

        # Normal cases - Tables
        if isinstance(join_obj.left, Table):
            stmt = self.rewrite_from_table(stmt, join_obj.left)

        if isinstance(join_obj.right, Table):
            stmt = self.rewrite_from_table(stmt, join_obj.right)

        # SB - This happened when managing "lazy-joined" relationship
        # if isinstance(join_obj.right, Alias) and isinstance(join_obj.left, Subquery):  # Lazy joins
        #     stmt = self.analyze_from(stmt, join_obj.right)

        return stmt

    def analyze_from(self, stmt: Select, from_obj):
        """Analyze the FROMS of a Select to determine possible soft-delete rewritable tables."""
        if isinstance(from_obj, Table):
            return self.rewrite_from_table(stmt, from_obj)

        if isinstance(from_obj, _ORMJoin) or isinstance(from_obj, Join):
            # _ORMJOIN/Join contains information about two things: 'left' and 'right'. Check both.
            return self.rewrite_from_orm_join(stmt, from_obj)

        if isinstance(from_obj, Subquery):
            self.rewrite_element(from_obj)
            return stmt

        if isinstance(from_obj, TableClause) or isinstance(from_obj, TextClause):
            # TableClause/TextClause objects are raw text SQL identifiers and as such, we cannot
            # introspect or do anything about this statement.
            return stmt

        if isinstance(from_obj, Alias):
            # SB - This happened when managing "lazy-joined" relationship
            # if isinstance(from_obj.element, Table):
            #     return self.rewrite_from_table(from_obj.element.select(), from_obj)
            #     # return self.rewrite_element(from_obj.element)

            if isinstance(from_obj.element, Subquery):
                self.rewrite_element(from_obj.element)
                return stmt

            raise NotImplementedError(
                f"Unsupported object \"{(type(from_obj.element))}\" inside Alias in " f"statement.froms"
            )

        raise NotImplementedError(f"Unsupported object \"{(type(from_obj))}\" in statement.froms")

    def rewrite_from_table(self, stmt: Select, table: Table) -> Select:
        """(possibly) Rewrite a Select based on whether the Table contains the soft-delete field or not."""
        column_obj = table.columns.get(self.deleted_field_name)

        # Caveat: The automatic "bool(column_obj)" conversion actually returns
        # a truthy value of False (?), so we have to explicitly compare against None
        if column_obj is not None:
            return stmt.filter(column_obj.is_(None))

        # Soft-delete argument was not found, return unchanged statement
        return stmt



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ParticipantEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16ParticipantEvent.proto\x12\x11opentera.protobuf\"\x92\x03\n\x10ParticipantEvent\x12\x18\n\x10participant_uuid\x18\x01 \x01(\t\x12;\n\x04type\x18\x02 \x01(\x0e\x32-.opentera.protobuf.ParticipantEvent.EventType\x12\x18\n\x10participant_name\x18\x03 \x01(\t\x12 \n\x18participant_project_name\x18\x04 \x01(\t\x12\x1d\n\x15participant_site_name\x18\x05 \x01(\t\"\xcb\x01\n\tEventType\x12\x17\n\x13PARTICIPANT_UNKNOWN\x10\x00\x12\x19\n\x15PARTICIPANT_CONNECTED\x10\x01\x12\x1c\n\x18PARTICIPANT_DISCONNECTED\x10\x02\x12\x17\n\x13PARTICIPANT_DELETED\x10\x03\x12\x15\n\x11PARTICIPANT_ADDED\x10\x04\x12\x1e\n\x1aPARTICIPANT_JOINED_SESSION\x10\x05\x12\x1c\n\x18PARTICIPANT_LEFT_SESSION\x10\x06\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ParticipantEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PARTICIPANTEVENT._serialized_start=46
  _PARTICIPANTEVENT._serialized_end=448
  _PARTICIPANTEVENT_EVENTTYPE._serialized_start=245
  _PARTICIPANTEVENT_EVENTTYPE._serialized_end=448
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: LogEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eLogEvent.proto\x12\x11opentera.protobuf\"\xa6\x02\n\x08LogEvent\x12\x33\n\x05level\x18\x01 \x01(\x0e\x32$.opentera.protobuf.LogEvent.LogLevel\x12\x11\n\ttimestamp\x18\x02 \x01(\x01\x12\x0e\n\x06sender\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"\xb0\x01\n\x08LogLevel\x12\x14\n\x10LOGLEVEL_UNKNOWN\x10\x00\x12\x12\n\x0eLOGLEVEL_TRACE\x10\x07\x12\x12\n\x0eLOGLEVEL_DEBUG\x10\x06\x12\x11\n\rLOGLEVEL_INFO\x10\x05\x12\x14\n\x10LOGLEVEL_WARNING\x10\x04\x12\x15\n\x11LOGLEVEL_CRITICAL\x10\x03\x12\x12\n\x0eLOGLEVEL_ERROR\x10\x02\x12\x12\n\x0eLOGLEVEL_FATAL\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'LogEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LOGEVENT._serialized_start=38
  _LOGEVENT._serialized_end=332
  _LOGEVENT_LOGLEVEL._serialized_start=156
  _LOGEVENT_LOGLEVEL._serialized_end=332
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TeraEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fTeraEvent.proto\x12\x11opentera.protobuf\x1a\x19google/protobuf/any.proto\"\x9e\x01\n\tTeraEvent\x12\x33\n\x06header\x18\x01 \x01(\x0b\x32#.opentera.protobuf.TeraEvent.Header\x12$\n\x06\x65vents\x18\x02 \x03(\x0b\x32\x14.google.protobuf.Any\x1a\x36\n\x06Header\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0c\n\x04time\x18\x02 \x01(\x01\x12\r\n\x05topic\x18\x03 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'TeraEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TERAEVENT._serialized_start=66
  _TERAEVENT._serialized_end=224
  _TERAEVENT_HEADER._serialized_start=170
  _TERAEVENT_HEADER._serialized_end=224
# @@protoc_insertion_point(module_scope)



import uuid
from flask import request
from requests import Response
from typing import List


class TeraUserClient:

    def __init__(self, token_dict: dict, token: str, config_man):
        self.__user_uuid = token_dict['user_uuid']
        self.__id_user = token_dict['id_user']
        self.__user_fullname = token_dict['user_fullname']
        self.__user_token = token
        self.__user_superadmin = token_dict['user_superadmin']
        self.__service_access = token_dict['service_access']

        backend_hostname = config_man.backend_config["hostname"]
        backend_port = str(config_man.backend_config["port"])

        self.__backend_url = 'https://' + backend_hostname + ':' + backend_port

    @property
    def user_uuid(self):
        return self.__user_uuid

    @user_uuid.setter
    def user_uuid(self, u_uuid: uuid):
        self.__user_uuid = u_uuid

    @property
    def id_user(self):
        return self.__id_user

    @id_user.setter
    def id_user(self, id_user: int):
        self.__id_user = id_user

    @property
    def user_fullname(self):
        return self.__user_fullname

    @user_fullname.setter
    def user_fullname(self, name: str):
        self.__user_fullname = name

    @property
    def user_token(self):
        return self.__user_token

    @user_token.setter
    def user_token(self, token: str):
        self.__user_token = token

    @property
    def user_superadmin(self):
        return self.__user_superadmin

    @user_superadmin.setter
    def user_superadmin(self, superadmin: bool):
        self.__user_superadmin = superadmin

    @property
    def service_access(self):
        return self.__service_access

    @service_access.setter
    def service_access(self, service_access: dict):
        self.__service_access = service_access

    def do_get_request_to_backend(self, path: str) -> Response:
        from requests import get
        request_headers = {'Authorization': 'OpenTera ' + self.__user_token}
        # TODO: remove verify=False and check certificate
        backend_response = get(url=self.__backend_url + path, headers=request_headers, verify=False)
        return backend_response

    def get_roles_for_service(self, service_key: str) -> List[str]:
        # Roles are stored in the token, in the service_access dictionary
        roles: List[str] = []

        if service_key in self.__service_access:
            roles = self.__service_access[service_key]
        return roles

    def get_role_for_site(self, id_site: int) -> str:
        response = self.do_get_request_to_backend('/api/user/sites?user_uuid=' + self.__user_uuid)

        if response.status_code == 200:
            # Parse JSON reply
            import json
            sites = json.loads(response.text)

            # Find correct site in reply
            for site in sites:
                if site['id_site'] == id_site:
                    return site['site_role']

        return 'Undefined'

    def get_role_for_project(self, id_project: int) -> str:
        response = self.do_get_request_to_backend('/api/user/projects?user_uuid=' + self.__user_uuid)

        if response.status_code == 200:
            # Parse JSON reply
            import json
            projects = json.loads(response.text)

            # Find correct site in reply
            for project in projects:
                if project['id_project'] == id_project:
                    return project['project_role']

        return 'Undefined'

    def get_user_info(self):
        response = self.do_get_request_to_backend('/api/user/users?user_uuid=' + self.__user_uuid)

        if response.status_code == 200:
            return response.json()[0]

        return {}

    def __repr__(self):
        return '<TeraUserClient - UUID: ' + self.__user_uuid + ', Token: ' + self.__user_token + '>'



from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from enum import Enum

import uuid
import json
from datetime import datetime


class TeraTestStatus(Enum):
    TEST_INCOMPLETE = 0
    TEST_INPROGRESS = 1
    TEST_COMPLETED = 2


class TeraTest(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_tests'
    id_test = Column(Integer, Sequence('id_test_sequence'), primary_key=True, autoincrement=True)
    id_test_type = Column(Integer, ForeignKey('t_tests_types.id_test_type'), nullable=False)
    id_session = Column(Integer, ForeignKey('t_sessions.id_session', ondelete='cascade'), nullable=False)

    # Item to which that test is associated. At least one of it should be selected, but multiple can also be specified
    # Of course, that item needs to be in the associated session for data integrity
    id_device = Column(Integer, ForeignKey("t_devices.id_device"), nullable=True)
    id_participant = Column(Integer, ForeignKey("t_participants.id_participant"), nullable=True)
    id_user = Column(Integer, ForeignKey("t_users.id_user"), nullable=True)
    id_service = Column(Integer, ForeignKey("t_services.id_service"), nullable=True)

    test_uuid = Column(String(36), nullable=False, unique=True)
    test_name = Column(String, nullable=False)
    test_datetime = Column(TIMESTAMP(timezone=True), nullable=False)
    test_status = Column(Integer, nullable=False, default=0)
    test_summary = Column(String, nullable=True)  # This contains a json formatted summary for results display

    test_session = relationship("TeraSession", back_populates='session_tests')
    test_device = relationship("TeraDevice", back_populates='device_tests')
    test_user = relationship("TeraUser", back_populates='user_tests')
    test_participant = relationship("TeraParticipant", back_populates='participant_tests')
    test_service = relationship("TeraService", back_populates='service_tests')
    test_test_type = relationship('TeraTestType', back_populates='test_type_tests')

    def from_json(self, json_data, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['test_session', 'test_device', 'test_user', 'test_participant', 'test_service',
                              'test_test_type'])

        super().from_json(json_data, ignore_fields)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['test_session', 'test_device', 'test_user', 'test_participant', 'test_service',
                              'test_test_type'])

        test_json = super().to_json(ignore_fields=ignore_fields)
        if not minimal:
            test_json['test_session_name'] = self.test_session.session_name
            test_json['test_test_type_name'] = self.test_test_type.test_type_name
            test_json['test_test_type_uuid'] = self.test_test_type.test_type_uuid
            if self.id_device:
                test_json['test_device'] = self.test_device.device_name
            if self.id_user:
                test_json['test_user'] = self.test_user.get_fullname()
            if self.id_participant:
                test_json['test_participant'] = self.test_participant.participant_name
            if self.id_service:
                test_json['test_service'] = self.test_service.service_name

        return test_json

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_test': self.id_test, 'test_uuid': self.test_uuid}

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSession import TeraSession
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraParticipant import TeraParticipant
            from opentera.db.models.TeraUser import TeraUser
            from opentera.db.models.TeraTestType import TeraTestType

            session2 = TeraSession.get_session_by_name("Séance #2")
            session3 = TeraSession.get_session_by_name("Séance #3")

            pretesttype = TeraTestType.get_test_type_by_id(1)
            posttesttype = TeraTestType.get_test_type_by_id(2)

            for i in range(4):
                new_test = TeraTest()
                new_test.test_name = "Test #" + str(i)
                new_test.test_session = session2
                new_test.test_uuid = str(uuid.uuid4())
                new_test.test_datetime = session2.session_start_datetime
                if i == 0:
                    new_test.id_test_type = pretesttype.id_test_type
                    new_test.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
                if i == 1:
                    new_test.id_test_type = posttesttype.id_test_type
                    new_test.id_user = TeraUser.get_user_by_id(2).id_user
                if i == 2:
                    new_test.id_test_type = pretesttype.id_test_type
                    new_test.id_device = TeraDevice.get_device_by_id(1).id_device
                if i == 3:
                    new_test.test_session = session3
                    new_test.test_datetime = session3.session_start_datetime
                    new_test.id_test_type = pretesttype.id_test_type
                    new_test.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
                TeraTest.db().session.add(new_test)

            TeraTest.db().session.commit()

    @staticmethod
    def get_test_by_id(test_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted).filter_by(id_test=test_id).first()

    @staticmethod
    def get_test_by_uuid(test_uuid: str, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted).filter_by(test_uuid=test_uuid).first()

    @staticmethod
    def get_tests_for_device(device_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device=device_id).order_by(TeraTest.id_test_type.asc(), TeraTest.test_datetime.asc()).all()

    @staticmethod
    def get_tests_for_user(user_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_user=user_id).order_by(TeraTest.id_test_type.asc(), TeraTest.test_datetime.asc()).all()

    @staticmethod
    def get_tests_for_session(session_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session=session_id).order_by(TeraTest.id_test_type.asc(), TeraTest.test_datetime.asc()).all()

    @staticmethod
    def get_tests_for_participant(part_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_participant=part_id).order_by(TeraTest.id_test_type.asc(), TeraTest.test_datetime.asc()).all()

    @staticmethod
    def get_tests_for_service(service_id: int, with_deleted: bool = False):
        return TeraTest.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id).order_by(TeraTest.id_test_type.asc(), TeraTest.test_datetime.asc()).all()

    @staticmethod
    def get_access_token(test_uuids: list, token_key: str, requester_uuid: str, expiration=3600):
        import time
        import jwt

        # Creating token with user info
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'test_uuids': test_uuids,
            'requester_uuid': requester_uuid
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    def get_service_url(self, server_url: str, server_port: int) -> dict:
        urls = {'test_answers_url': None,
                'test_answers_web_url': None,
                }

        if not self.test_test_type.test_type_service.service_enabled:
            return urls  # Service disabled = no Urls!

        service_endpoint = self.test_test_type.test_type_service.service_clientendpoint
        base_url = 'https://' + server_url + ':' + str(server_port) + service_endpoint

        urls['test_answers_url'] = base_url + '/api/tests/answers'

        if self.test_test_type.test_type_has_web_format:
            urls['test_answers_web_url'] = base_url + '/api/tests/answers/web'

        return urls

    @classmethod
    def insert(cls, test):
        # Generate UUID
        test.test_uuid = str(uuid.uuid4())

        # Check if summary is in json
        if isinstance(test.test_summary, dict):
            test.test_summary = json.dumps(test.test_summary)

        if not test.test_datetime:
            test.test_datetime = datetime.now()

        super().insert(test)

    @classmethod
    def update(cls, update_id: int, values: dict):

        # Check if summary is in json
        if 'test_summary' in values:
            if isinstance(values['test_summary'], dict):
                values['test_summary'] = json.dumps(values['test_summary'])

        super().update(update_id=update_id, values=values)



from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraUserGroupForm:

    @staticmethod
    def get_user_group_form():
        form = TeraForm("user_group")

        # Building lists
        #################
        # None to build!

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_user_group", gettext("User Group ID"), "hidden",
                                      item_required=True))
        section.add_item(TeraFormItem("user_group_name", gettext("User Group Name"), "text",
                                      item_required=True))

        return form.to_dict()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraSessionParticipants(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_sessions_participants'
    id_session_participant = Column(Integer, Sequence('id_session_participant'), primary_key=True, autoincrement=True)
    id_session = Column(Integer, ForeignKey('t_sessions.id_session', ondelete='cascade'))
    id_participant = Column(Integer, ForeignKey('t_participants.id_participant'))

    session_participant_session = relationship('TeraSession', viewonly=True)
    session_participant_participant = relationship('TeraParticipant', viewonly=True)

    @staticmethod
    def get_session_count_for_participant(id_participant: int, with_deleted: bool = False) -> int:
        return TeraSessionParticipants.get_count(filters={'id_participant': id_participant}, with_deleted=with_deleted)

    @classmethod
    def update(cls, update_id: int, values: dict):
        return




from flask import request
from ua_parser import user_agent_parser


class UserAgentParser:
    @staticmethod
    def parse_request_for_login_infos(http_request: request) -> dict:
        infos = {'client_name': '',
                 'client_version': '',
                 'os_name': '',
                 'os_version': '',
                 'server_endpoint': http_request.endpoint,
                 'client_ip': http_request.remote_addr
                 }

        if 'USER_AGENT' in http_request.headers:
            user_agent = user_agent_parser.Parse(http_request.headers['USER_AGENT'])
            infos['client_name'] = ''
            if user_agent['device']['brand'] and user_agent['device']['brand'] != 'Other':
                infos['client_name'] = user_agent['device']['brand'] + ' - '
            if user_agent['device']['family'] and user_agent['device']['family'] != 'Other':
                infos['client_name'] += user_agent['device']['family'] + ' - '

            infos['client_name'] += user_agent['user_agent']['family']

            if user_agent['user_agent']['major']:
                infos['client_version'] = user_agent['user_agent']['major']
            if user_agent['user_agent']['minor']:
                infos['client_version'] += '.' + user_agent['user_agent']['minor']
            if user_agent['user_agent']['patch']:
                infos['client_version'] += '.' + user_agent['user_agent']['patch']

            infos['os_name'] = user_agent['os']['family']

            if user_agent['os']['major']:
                infos['os_version'] = user_agent['os']['major']
            if user_agent['os']['minor']:
                infos['os_version'] += '.' + user_agent['os']['minor']
            if user_agent['os']['patch']:
                infos['os_version'] += '.' + user_agent['os']['patch']

        if 'X-Client-Name' in http_request.headers:
            infos['client_name'] = http_request.headers['X-Client-Name']
        if 'X-Client-Version' in http_request.headers:
            infos['client_version'] = http_request.headers['X-Client-Version']
        if 'X-OS-Name' in http_request.headers:
            infos['os_name'] = http_request.headers['X-OS-Name']
        if 'X-OS-Version' in http_request.headers:
            infos['os_version'] = http_request.headers['X-OS-Version']

        if 'X-Script-Name' in http_request.headers:
            infos['server_endpoint'] = http_request.headers['X-Script-Name']

        return infos



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: DeviceEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x44\x65viceEvent.proto\x12\x11opentera.protobuf\"\xcc\x02\n\x0b\x44\x65viceEvent\x12\x13\n\x0b\x64\x65vice_uuid\x18\x01 \x01(\t\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32(.opentera.protobuf.DeviceEvent.EventType\x12\x13\n\x0b\x64\x65vice_name\x18\x03 \x01(\t\x12\x15\n\rdevice_status\x18\x04 \x01(\t\"\xc3\x01\n\tEventType\x12\x12\n\x0e\x44\x45VICE_UNKNOWN\x10\x00\x12\x14\n\x10\x44\x45VICE_CONNECTED\x10\x01\x12\x17\n\x13\x44\x45VICE_DISCONNECTED\x10\x02\x12\x12\n\x0e\x44\x45VICE_DELETED\x10\x03\x12\x10\n\x0c\x44\x45VICE_ADDED\x10\x04\x12\x19\n\x15\x44\x45VICE_JOINED_SESSION\x10\x05\x12\x17\n\x13\x44\x45VICE_LEFT_SESSION\x10\x06\x12\x19\n\x15\x44\x45VICE_STATUS_CHANGED\x10\x07\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'DeviceEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DEVICEEVENT._serialized_start=41
  _DEVICEEVENT._serialized_end=373
  _DEVICEEVENT_EVENTTYPE._serialized_start=178
  _DEVICEEVENT_EVENTTYPE._serialized_end=373
# @@protoc_insertion_point(module_scope)



from opentera.db.Base import BaseModel
from opentera.db.models.TeraSession import TeraSession
from opentera.db.models.TeraSessionType import TeraSessionType
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraSessionTypeProject(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_sessions_types_projects'
    id_session_type_project = Column(Integer, Sequence('id_session_type_project_sequence'), primary_key=True,
                                     autoincrement=True)
    id_session_type = Column('id_session_type', Integer, ForeignKey('t_sessions_types.id_session_type',
                                                                    ondelete='cascade'), nullable=False)
    id_project = Column('id_project', Integer, ForeignKey('t_projects.id_project', ondelete='cascade'),
                        nullable=False)

    session_type_project_session_type = relationship("TeraSessionType", viewonly=True)
    session_type_project_project = relationship("TeraProject", viewonly=True)

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['session_type_project_session_type', 'session_type_project_project'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    def to_json_create_event(self):
        json_data = self.to_json(minimal=True)
        # Query session type information
        session_type = TeraSessionType.get_session_type_by_id(self.id_session_type)
        if session_type:
            json_data['session_type_id_service'] = session_type.id_service
        return json_data

    def to_json_update_event(self):
        json_data = self.to_json(minimal=True)
        json_data['session_type_id_service'] = self.session_type_project_session_type.id_service
        return json_data

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_session_type_project': self.id_session_type_project}

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSessionType import TeraSessionType
            from opentera.db.models.TeraProject import TeraProject

            video_session = TeraSessionType.get_session_type_by_id(1)
            sensor_session = TeraSessionType.get_session_type_by_id(2)
            data_session = TeraSessionType.get_session_type_by_id(3)
            exerc_session = TeraSessionType.get_session_type_by_id(4)
            bureau_session = TeraSessionType.get_session_type_by_id(5)

            project = TeraProject.get_project_by_projectname('Default Project #1')

            stp = TeraSessionTypeProject()
            stp.id_session_type = video_session.id_session_type
            stp.id_project = project.id_project
            TeraSessionTypeProject.db().session.add(stp)

            stp = TeraSessionTypeProject()
            stp.id_session_type = sensor_session.id_session_type
            stp.id_project = project.id_project
            TeraSessionTypeProject.db().session.add(stp)

            stp = TeraSessionTypeProject()
            stp.id_session_type = data_session.id_session_type
            stp.id_project = project.id_project
            TeraSessionTypeProject.db().session.add(stp)

            stp = TeraSessionTypeProject()
            stp.id_session_type = exerc_session.id_session_type
            stp.id_project = project.id_project
            TeraSessionTypeProject.db().session.add(stp)

            stp = TeraSessionTypeProject()
            stp.id_session_type = bureau_session.id_session_type
            stp.id_project = project.id_project
            TeraSessionTypeProject.db().session.add(stp)

            TeraSessionTypeProject.db().session.commit()

    @staticmethod
    def get_session_type_project_by_id(stp_id: int, with_deleted: bool = False):
        return TeraSessionTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_type_project=stp_id).first()

    @staticmethod
    def get_projects_for_session_type(session_type_id: int, with_deleted: bool = False):
        return TeraSessionTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_type=session_type_id).all()

    @staticmethod
    def get_sessions_types_for_project(project_id: int, with_deleted: bool = False):
        return TeraSessionTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id).all()

    @staticmethod
    def get_session_type_project_for_session_type_project(project_id: int, session_type_id: int,
                                                          with_deleted: bool = False):
        return TeraSessionTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id, id_session_type=session_type_id).first()

    @staticmethod
    def get_session_type_project_for_project_and_service(project_id: int, service_id: int, with_deleted: bool = False):
        from opentera.db.models.TeraSessionType import TeraSessionType
        return TeraSessionTypeProject.query.execution_options(include_deleted=with_deleted).join(TeraSessionType).\
            filter(TeraSessionType.id_service == service_id).\
            filter(TeraSessionTypeProject.id_project == project_id).all()

    @staticmethod
    def delete_with_ids(session_type_id: int, project_id: int, autocommit: bool = True):
        delete_obj: TeraSessionTypeProject = TeraSessionTypeProject.query.filter_by(id_session_type=session_type_id,
                                                                                    id_project=project_id).first()
        if delete_obj:
            TeraSessionTypeProject.delete(delete_obj.id_session_type_project, autocommit=autocommit)

    @staticmethod
    def check_integrity(obj_to_check):
        from opentera.db.models.TeraSessionType import TeraSessionType
        # If that session type is related to a service, make sure that the service is associated to that project
        if obj_to_check.session_type_project_session_type.session_type_category == \
                TeraSessionType.SessionCategoryEnum.SERVICE.value:
            service_projects = [proj.id_project for proj in
                                obj_to_check.session_type_project_session_type.session_type_service.service_projects]
            if obj_to_check.id_project not in service_projects:
                # We must also associate that service to that project!
                from opentera.db.models.TeraServiceProject import TeraServiceProject
                new_service_project = TeraServiceProject()
                new_service_project.id_service = obj_to_check.session_type_project_session_type \
                    .session_type_service.id_service
                new_service_project.id_project = obj_to_check.session_type_project_project.id_project
                TeraServiceProject.insert(new_service_project)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        sessions = TeraSession.get_sessions_for_project(project_id=self.id_project,
                                                        session_type_id=self.id_session_type,
                                                        with_deleted=with_deleted)
        if len(sessions) > 0:
            return IntegrityError('Session type has sessions in this project', self.id_session_type, 't_sessions')
        return None

    @classmethod
    def insert(cls, stp):
        # Check if that site of that project has the site associated to it
        from opentera.db.models.TeraSessionTypeSite import TeraSessionTypeSite
        from opentera.db.models.TeraProject import TeraProject
        project = TeraProject.get_project_by_id(project_id=stp.id_project)
        st_site = TeraSessionTypeSite.get_session_type_site_for_session_type_and_site(site_id=project.id_site,
                                                                                      session_type_id=
                                                                                      stp.id_session_type)
        if not st_site:
            raise IntegrityError(params='Session type not associated to project site',
                                 orig='TeraSessionTypeProject.insert', statement='insert')
        inserted_obj = super().insert(stp)
        TeraSessionTypeProject.check_integrity(inserted_obj)
        return inserted_obj

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.utils.TeraVersions import TeraVersions


class TeraVersionsForm:

    @staticmethod
    def get_versions_form():
        form = TeraForm("versions")

        # Building lists
        #################
        # None to build!

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("version_string", gettext("OpenTeraServer version string"), "hidden",
                                      item_required=True, item_options={"readonly": True}))

        section.add_item(TeraFormItem("version_major", gettext("OpenTeraServer major version number"), "text",
                                      item_required=True, item_options={"readonly": True}))

        section.add_item(TeraFormItem("version_minor", gettext("OpenTeraServer minor version number"), "text",
                                      item_required=True, item_options={"readonly": True}))

        section.add_item(TeraFormItem("version_patch", gettext("OpenTeraServer patch version number"), "text",
                                      item_required=True, item_options={"readonly": True}))

        # Load versions from DB
        versions = TeraVersions()
        versions.load_from_db()
        versions_dict = versions.to_dict()

        # One section per client.
        for client_string in versions_dict['clients']:
            # Create a section for each client
            client = versions_dict['clients'][client_string]
            section_clients = TeraFormSection(client['client_name'], client['client_name'] + ' ' + gettext("Versions"))

            section_clients.add_item(TeraFormItem("client_name", gettext("Client name"), "text", item_required=True,
                                                  item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_description", gettext("Client description"), "longtext",
                                                  item_required=True, item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_version", gettext("Client version"), "text",
                                                  item_required=True, item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_documentation_url", gettext("Client documentation url"),
                                                  "text", item_required=True, item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_windows_download_url", gettext("Client windows version"),
                                                  "text", item_required=True, item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_mac_download_url", gettext("Client mac version"), "text",
                                                  item_required=True, item_options={"readonly": True}))

            section_clients.add_item(TeraFormItem("client_linux_download_url", gettext("Client linux version"), "text",
                                                  item_required=True, item_options={"readonly": True}))

            form.add_section(section_clients)

        return form.to_dict()



import redis
from opentera.messages.python.RPCMessage_pb2 import RPCMessage, Value
from datetime import datetime
import json
import uuid
from opentera.redis.RedisVars import RedisVars


class RedisRPCClient:
    def __init__(self, config: dict, timeout=10):
        self.config = config
        self.timeout = timeout
        # Create a unique random pattern
        self.pattern = 'RedisRPCClient.' + str(uuid.uuid4())
        self.msg_id = 0
        self.client = redis.Redis(host=self.config['hostname'],
                                  port=self.config['port'],
                                  db=self.config['db'],
                                  username=self.config['username'],
                                  password=self.config['password'],
                                  client_name=self.pattern)

    # def __del__(self):
    #     print('****- Deleting RedisRPCClient')

    def _internal_rpc_call(self, topic: str, function_name: str, *args):
        # Pub/Sub client
        p = self.client.pubsub()

        message = RPCMessage()
        message.method = function_name
        message.timestamp = datetime.now().timestamp()
        message.id = self.msg_id
        self.msg_id = self.msg_id + 1
        message.reply_to = self.pattern

        # Iterate through args
        rpc_args = []
        for arg in args:
            if isinstance(arg, bool):
                val = Value()
                val.bool_value = arg
                rpc_args.append(val)
            elif isinstance(arg, float):
                val = Value()
                val.double_value = arg
                rpc_args.append(val)
            elif isinstance(arg, int):
                val = Value()
                val.int_value = arg
                rpc_args.append(val)
            elif isinstance(arg, str):
                val = Value()
                val.string_value = arg
                rpc_args.append(val)
            elif isinstance(arg, bytes):
                val = Value()
                val.bytes_value = arg
                rpc_args.append(val)
            else:
                print('Invalid arg:', arg)

        # Set args
        message.args.extend(rpc_args)

        # Will answer on the reply_to field
        p.subscribe(message.reply_to)
        # First message is for subscribe result
        message1 = p.get_message(timeout=self.timeout)

        # Publish request
        self.client.publish(topic, message.SerializeToString())

        # Second message is data received
        message2 = p.get_message(timeout=self.timeout)

        # Unsubscribe to message
        # p.unsubscribe(message.reply_to)
        # Closing (will automatically unsubscribe)
        p.close()

        if message2:
            result = json.loads(message2['data'])
            return result['return_value']

        return None

    def call(self, module_name: str, function_name: str, *args):
        return self._internal_rpc_call('module.' + module_name + '.rpc', function_name, *args)

    def call_service(self, service_key: str, function_name: str, *args):
        return self._internal_rpc_call(RedisVars.build_service_rpc_topic(service_key), function_name, *args)




from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraDeviceSubTypeForm:

    @staticmethod
    def get_device_subtype_form():
        form = TeraForm("device_subtype")

        # Building lists
        #################
        # Device types
        from opentera.db.models.TeraDeviceType import TeraDeviceType
        device_types = TeraDeviceType.get_devices_types()
        device_types_list = []
        for dev_type in device_types:
            name = gettext(dev_type.get_name())
            device_types_list.append(TeraFormValue(value_id=dev_type.id_device_type, value=name))

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_device_subtype", gettext("Device Sub-Type ID"), "hidden",
                                      item_required=True))
        section.add_item(TeraFormItem("device_subtype_name", gettext("Device Sub-Type Name"), "text",
                                      item_required=True))
        section.add_item(TeraFormItem("id_device_type", gettext("Device Type ID"), "array", item_required=True,
                                      item_values=device_types_list))

        return form.to_dict()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("env/requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.readlines()
    requirements.append('protobuf==3.20.3')

setuptools.setup(
    name="opentera",
    version="1.2.5",
    author="Dominic Létourneau, Simon Brière",
    author_email="dominic.letourneau@usherbrooke.ca, simon.briere@usherbrooke.ca",
    description="OpenTera base package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/introlab/opentera",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)



# Event based, twisted redis
import txredisapi as txredis
from twisted.internet import defer


class redisProtocol(txredis.SubscriberProtocol):
    def __init__(self, charset=None, errors="strict", parent=None, *args, **kwargs):
        # print('redisProtocol init with args', charset, errors, parent, args, kwargs)
        super().__init__(charset, errors, *args, **kwargs)
        self.parent = parent
        if self.parent:
            self.parent.setProtocol(self)

    # def __del__(self):
    #     print("****- Deleting redisProtocol")

    @defer.inlineCallbacks
    def connectionMade(self):
        # print('redisProtocol connectionMade')

        # Call base class connectionMade to handle password and other init stuff
        super().connectionMade()

        if self.parent:
            ret = yield self.execute_command('client', 'setname', 'txredis_' + self.parent.__class__.__name__)
            self.parent.redisConnectionMade()
        else:
            print('redisProtocol connectionMade')

    def messageReceived(self, pattern, channel, message):
        # print('redisProtocol message received', pattern, channel, message)
        if self.parent:
            self.parent.redisMessageReceived(pattern, channel, message)
        else:
            print('redisProtocol message received', pattern, channel, message)

    def connectionLost(self, reason):
        # print("redisProtocol lost connection", reason)

        # Call base class connectionLost
        super().connectionLost(reason)

        if self.parent:
            self.parent.redisConnectionLost(reason)
        else:
            print("redisProtocol lost connection", reason)

    def replyReceived(self, reply):
        super().replyReceived(reply)


class RedisProtocolFactory(txredis.SubscriberFactory):

    maxDelay = 120
    continueTrying = True
    protocol = None

    def __init__(self, parent, protocol, config):
        super().__init__()
        # print('setting arg', parent)
        # print('setting protocol', protocol)
        self.parent = parent
        self.protocol = protocol
        self.redis_config = config

    # def __del__(self):
    #     print("****- Deleting RedisProtocolFactory")

    def buildProtocol(self, addr):
        """
        This overloaded function is called when a new protocol (effective connection to the redis server) is
        performed.
        :param addr:
        :return:
        """
        # print('build Protocol addr:', addr)
        if hasattr(self, 'charset'):
            # p = self.protocol(parent=self.parent)
            p = self.protocol(self.charset,
                              parent=self.parent,
                              password=None if self.redis_config['password'] == '' else self.redis_config['password'],
                              dbid=self.redis_config['db'])
        else:
            # Forcing no encoding
            p = self.protocol(parent=self.parent,
                              password=self.redis_config['password'],
                              dbid=self.redis_config['db'])
        p.factory = self
        p.timeOut = self.replyTimeout
        return p



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess


class TeraUserForm:

    @staticmethod
    def get_user_form():
        form = TeraForm("user")

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_user", gettext("User ID"), "hidden", True))
        section.add_item(TeraFormItem("user_uuid", gettext("User UUID"), "hidden"))
        section.add_item(TeraFormItem("user_name", gettext("User Full Name"), "hidden"))
        section.add_item(TeraFormItem("user_username", gettext("Username"), "text", True))
        section.add_item(TeraFormItem("user_enabled", gettext("User Enabled"), "boolean", True, item_default=True))
        section.add_item(TeraFormItem("user_firstname", gettext("First Name"), "text", True))
        section.add_item(TeraFormItem("user_lastname", gettext("Last Name"), "text", True))
        section.add_item(TeraFormItem("user_email", gettext("Email"), "text"))
        section.add_item(
            TeraFormItem("user_password", gettext("Password"), "password", item_options={"confirm": True}))
        section.add_item(TeraFormItem("user_superadmin", gettext("User Is Super Administrator"), "boolean", True))
        section.add_item(TeraFormItem("user_notes", gettext("Notes"), "longtext"))
        section.add_item(TeraFormItem("user_profile", gettext("Profile"), "hidden"))
        section.add_item(TeraFormItem("user_lastonline", gettext("Last Connection"), "datetime",
                                      item_options={"readonly": True}))

        return form.to_dict()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: DatabaseEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x44\x61tabaseEvent.proto\x12\x11opentera.protobuf\"\xbe\x01\n\rDatabaseEvent\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32*.opentera.protobuf.DatabaseEvent.EventType\x12\x13\n\x0bobject_type\x18\x02 \x01(\t\x12\x14\n\x0cobject_value\x18\x03 \x01(\t\"H\n\tEventType\x12\x0e\n\nDB_UNKNOWN\x10\x00\x12\r\n\tDB_CREATE\x10\x01\x12\r\n\tDB_UPDATE\x10\x02\x12\r\n\tDB_DELETE\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'DatabaseEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DATABASEEVENT._serialized_start=43
  _DATABASEEVENT._serialized_end=233
  _DATABASEEVENT_EVENTTYPE._serialized_start=161
  _DATABASEEVENT_EVENTTYPE._serialized_end=233
# @@protoc_insertion_point(module_scope)



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from opentera.db.models.TeraServiceRole import TeraServiceRole


class TeraServiceAccess(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_services_access'
    id_service_access = Column(Integer, Sequence('id_service_project_role_sequence'), primary_key=True,
                               autoincrement=True)
    id_user_group = Column(Integer, ForeignKey('t_users_groups.id_user_group', ondelete='cascade'),
                           nullable=True)
    id_device = Column(Integer, ForeignKey('t_devices.id_device', ondelete='cascade'), nullable=True)
    id_participant_group = Column(Integer, ForeignKey('t_participants_groups.id_participant_group', ondelete='cascade'),
                                  nullable=True)
    id_service_role = Column(Integer, ForeignKey('t_services_roles.id_service_role', ondelete='cascade'),
                             nullable=False)

    service_access_role = relationship("TeraServiceRole", viewonly=True)
    service_access_user_group = relationship("TeraUserGroup", viewonly=True)
    service_access_device = relationship("TeraDevice", viewonly=True)
    service_access_participant_group = relationship("TeraParticipantGroup", viewonly=True)

    def __init__(self):
        pass

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_access_role', 'service_access_user_group',
                              'service_access_device', 'service_access_participant_group', 'id_service_role'])

        if minimal:
            ignore_fields.extend([])

        json_val = super().to_json(ignore_fields=ignore_fields)

        # Also expands with service_role infos
        if self.service_access_role:
            json_val.update(self.service_access_role.to_json(minimal=minimal))

        # Remove null values
        if not json_val['id_device']:
            del json_val['id_device']
        if not json_val['id_participant_group']:
            del json_val['id_participant_group']
        if not json_val['id_user_group']:
            del json_val['id_user_group']

        # Complete information if not minimal
        if not minimal:
            if self.service_access_role:
                json_val['service_name'] = self.service_access_role.service_role_service.service_name
                json_val['service_key'] = self.service_access_role.service_role_service.service_key
                json_val['id_service'] = self.service_access_role.service_role_service.id_service
                json_val['service_access_role_name'] = self.service_access_role.service_role_name
            # else:
            #     # This happens on transient objects
            #     from opentera.db.models.TeraService import TeraService
            #     service = TeraService.get_service_by_id(self.id_service)
            #     if service:
            #         json_val['service_name'] = service.service_name
            if self.id_user_group:
                if self.service_access_user_group:
                    json_val['user_group_name'] = self.service_access_user_group.user_group_name
                else:
                    # This happens on transient objects
                    from opentera.db.models.TeraUserGroup import TeraUserGroup
                    ug = TeraUserGroup.get_user_group_by_id(self.id_user_group)
                    if ug:
                        json_val['user_group_name'] = ug.user_group_name
            if self.id_device:
                json_val['device_name'] = self.service_access_device.device_name
            if self.id_participant_group:
                json_val['participant_group_name'] = self.service_access_participant_group.participant_group_name
        return json_val

    @staticmethod
    def get_service_access_by_id(service_access_id: int, with_deleted: bool = False):
        return TeraServiceAccess.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_access=service_access_id).first()

    @staticmethod
    def update_service_access_for_user_group(id_service: int, id_user_group: int, id_service_role: int,
                                             id_site: int | None = None, id_project: int | None = None):
        # Check if access already exists
        access = TeraServiceAccess.get_service_access_for_user_group(id_service=id_service,
                                                                     id_user_group=id_user_group,
                                                                     id_site=id_site,
                                                                     id_project=id_project,
                                                                     id_service_role=id_service_role)

        if not access:
            # No access already present for that user group - create new one
            access = TeraServiceAccess()
            access.id_user_group = id_user_group
            access.id_service_role = id_service_role
            TeraServiceAccess.insert(access)
            return access
        else:
            # Update it
            access[0].id_service_role = id_service_role
            TeraServiceAccess.db().session.commit()
            return access[0]

    @staticmethod
    def get_service_access_for_user_group(id_user_group: int, id_service: int, id_project: int | None = None,
                                          id_site: int | None = None, id_service_role: int | None = None,
                                          with_deleted: bool = False):
        query = TeraServiceAccess.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_user_group=id_user_group).join(TeraServiceRole).filter_by(id_service=id_service)

        if id_project:
            query = query.filter_by(id_project=id_project)

        if id_site:
            query = query.filter_by(id_site=id_site)

        if id_service_role and not id_site and not id_project:  # We search with id_service_role only if no project/site
            query = query.filter_by(id_service_role=id_service_role)

        return query.all()

    @staticmethod
    # Returns all access related to a service
    def get_service_access(id_service: int, id_project: int | None = None, id_site: int | None = None,
                           with_deleted: bool = False):
        query = TeraServiceAccess.query.execution_options(include_deleted=with_deleted) \
            .join(TeraServiceRole).filter_by(id_service=id_service)\

        if id_site:
            query = query.filter_by(id_site=id_site)
        if id_project:
            query = query.filter_by(id_project=id_project)

        return query.all()

    @staticmethod
    def delete_service_access_for_user_group(id_service: int, id_user_group: int, id_site: int | None = None,
                                             id_project: int | None = None, id_service_role: int | None = None):

        for service_access in TeraServiceAccess.get_service_access_for_user_group(id_service=id_service,
                                                                                  id_user_group=id_user_group,
                                                                                  id_site=id_site,
                                                                                  id_project=id_project):
            if id_service_role and service_access.id_service_role == id_service_role:
                TeraServiceAccess.delete(service_access.id_service_access)
                return

            if id_site and service_access.service_access_role.id_site == id_site:
                TeraServiceAccess.delete(service_access.id_service_access)
                return

            if id_project and service_access.service_access_role.id_project == id_project:
                TeraServiceAccess.delete(service_access.id_service_access)
                return

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraUserGroup import TeraUserGroup
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraParticipantGroup import TeraParticipantGroup

            device = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            group = TeraParticipantGroup.get_participant_group_by_group_name('Default Participant Group A')

            service_bureau = TeraService.get_service_by_key('BureauActif')
            service_bureau_admin = service_bureau.service_roles[0]
            service_bureau_user = service_bureau.service_roles[1]

            service_logging = TeraService.get_service_by_key('LoggingService')
            service_logging_admin = service_logging.service_roles[0]
            service_logging_user = service_logging.service_roles[1]

            user_group1 = TeraUserGroup.get_user_group_by_group_name('Users - Project 1')
            user_group2 = TeraUserGroup.get_user_group_by_group_name('Admins - Project 1')

            service_role = TeraServiceAccess()
            service_role.id_user_group = user_group1.id_user_group
            service_role.id_service_role = service_bureau_admin.id_service_role
            TeraServiceAccess.db().session.add(service_role)

            service_role = TeraServiceAccess()
            service_role.id_user_group = user_group2.id_user_group
            service_role.id_service_role = service_logging_admin.id_service_role
            TeraServiceAccess.db().session.add(service_role)

            service_role = TeraServiceAccess()
            service_role.id_device = device.id_device
            service_role.id_service_role = service_bureau_user.id_service_role
            TeraServiceAccess.db().session.add(service_role)

            service_role = TeraServiceAccess()
            service_role.id_participant_group = group.id_participant_group
            service_role.id_service_role = service_logging_user.id_service_role
            TeraServiceAccess.db().session.add(service_role)

            TeraServiceAccess.db().session.commit()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship, lazyload
from sqlalchemy.exc import IntegrityError
from opentera.db.models.TeraParticipantGroup import TeraParticipantGroup
from opentera.db.models.TeraServerSettings import TeraServerSettings
from opentera.db.models.TeraSessionParticipants import TeraSessionParticipants
from opentera.db.models.TeraSession import TeraSession
from opentera.db.models.TeraAsset import TeraAsset
from opentera.db.models.TeraTest import TeraTest
from opentera.db.models.TeraProject import TeraProject

import uuid
import jwt
import datetime
from passlib.hash import bcrypt


# Generator for jti
def infinite_jti_sequence():
    num = 0
    while True:
        yield num
        num += 1


# Initialize generator, call next(participant_jti_generator) to get next sequence number
participant_jti_generator = infinite_jti_sequence()


class TeraParticipant(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_participants'
    id_participant = Column(Integer, Sequence('id_participant_sequence'), primary_key=True, autoincrement=True)
    participant_uuid = Column(String(36), nullable=False, unique=True)
    participant_name = Column(String, nullable=False)
    participant_username = Column(String(50), nullable=True)
    participant_email = Column(String, nullable=True)
    participant_password = Column(String, nullable=True)
    participant_token_enabled = Column(Boolean, nullable=False, default=False)
    participant_token = Column(String, nullable=True, unique=True)
    participant_lastonline = Column(TIMESTAMP(timezone=True), nullable=True)
    participant_enabled = Column(Boolean, nullable=False, default=True)
    participant_login_enabled = Column(Boolean, nullable=False, default=False)
    id_participant_group = Column(Integer, ForeignKey('t_participants_groups.id_participant_group', ondelete='cascade'),
                                  nullable=True)

    id_project = Column(Integer, ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)

    participant_devices = relationship("TeraDevice", secondary="t_devices_participants",
                                       back_populates="device_participants", viewonly=True)

    participant_sessions = relationship("TeraSession", secondary="t_sessions_participants",
                                        back_populates="session_participants", passive_deletes=True)

    participant_participant_group = relationship("TeraParticipantGroup",
                                                 back_populates='participant_group_participants')

    participant_project = relationship("TeraProject", back_populates='project_participants', lazy='selectin')

    participant_created_sessions = relationship("TeraSession", cascade='delete',
                                                back_populates='session_creator_participant', passive_deletes=True)

    participant_service_config = relationship("TeraServiceConfig", cascade='delete', passive_deletes=True)

    participant_assets = relationship("TeraAsset", cascade='delete', back_populates='asset_participant',
                                      passive_deletes=True)

    participant_tests = relationship("TeraTest", cascade='delete',
                                     back_populates='test_participant', passive_deletes=True)

    authenticated = False
    fullAccess = False

    def __init__(self):
        pass

    def __str__(self):
        return '<TeraParticipant ' + str(self.participant_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def dynamic_token(self, token_key: str, expiration=3600):
        import time
        import jwt

        # Creating token with participant info
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'jti': next(participant_jti_generator),
            'participant_uuid': self.participant_uuid,
            'id_participant': self.id_participant,
            'user_fullname': self.participant_name
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    def create_token(self):
        # Creating token with user info
        payload = {
            'iss': 'TeraServer',
            'jti': next(participant_jti_generator),
            'participant_uuid': self.participant_uuid,
            'id_participant': self.id_participant
        }

        self.participant_token = jwt.encode(payload, TeraServerSettings.get_server_setting_value(
            TeraServerSettings.ServerParticipantTokenKey), algorithm='HS256')

        return self.participant_token

    def update_last_online(self):
        self.participant_lastonline = datetime.datetime.now()
        TeraParticipant.db().session.commit()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['authenticated', 'participant_devices',
                              'participant_sessions', 'participant_password',
                              'participant_project', 'participant_participant_group', 'fullAccess',
                              'participant_created_sessions', 'participant_service_config'
                              ])
        if minimal:
            ignore_fields.extend(['participant_username', 'participant_lastonline',
                                  'participant_login_enabled', 'participant_token'])

        participant_json = super().to_json(ignore_fields=ignore_fields)
        if self.participant_project and not minimal:
            participant_json['participant_project_enabled'] = self.participant_project.project_enabled
        return participant_json

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_participant': self.id_participant, 'participant_uuid': self.participant_uuid}

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

    def is_active(self):
        return self.participant_enabled

    def is_login_enabled(self):
        return self.participant_login_enabled

    def get_id(self):
        return self.participant_uuid

    def get_first_session(self):
        session = (TeraSessionParticipants.query.filter_by(id_participant=self.id_participant)
                   .order_by(TeraSessionParticipants.id_session.asc()).limit(1).first())
        if session:
            # Turn off lazy loading for session
            return TeraSession.query.filter_by(id_session=session.id_session).options(lazyload("*")).first()
        # sessions = sorted(self.participant_sessions, key=lambda session: session.session_start_datetime)
        # if sessions:
        #     return sessions[0]
        return None

    def get_last_session(self):
        # from opentera.db.models.TeraSession import TeraSessionStatus
        # sessions = [session for session in self.participant_sessions
        #             if session.session_status == TeraSessionStatus.STATUS_COMPLETED.value or
        #             session.session_status == TeraSessionStatus.STATUS_TERMINATED.value]
        # sessions = sorted(sessions, key=lambda session: session.session_start_datetime)
        # if sessions:
        #     return sessions[-1]
        session = (TeraSessionParticipants.query.filter_by(id_participant=self.id_participant)
                   .order_by(TeraSessionParticipants.id_session.desc()).limit(1).first())
        if session:
            # Turn off lazy loading for session
            return TeraSession.query.filter_by(id_session=session.id_session).options(lazyload("*")).first()
        return None

    @staticmethod
    def encrypt_password(password):
        return bcrypt.hash(password)

    # @staticmethod
    # def is_anonymous():
    #     return False

    @staticmethod
    def verify_password(username, password, participant=None):
        # Query User with that username
        if participant is None:
            participant = TeraParticipant.get_participant_by_username(username)

        if participant is None:
            print('TeraParticipant: verify_password - participant ' + username + ' not found.')
            return None

        # Check if enabled
        if not participant.participant_enabled or not participant.participant_login_enabled:
            print('TeraUser: verify_password - user ' + username + ' is inactive or login is disabled.')
            return None

        # Check password
        if bcrypt.verify(password, participant.participant_password):
            participant.authenticated = True
            return participant

        return None

    @staticmethod
    def get_participant_by_token(token, with_deleted: bool = False):
        participant = TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_token=token).first()

        if participant and participant.participant_enabled and participant.participant_token_enabled:
            # Validate token
            data = jwt.decode(token.encode('utf-8'), TeraServerSettings.get_server_setting_value(
                TeraServerSettings.ServerParticipantTokenKey), algorithms='HS256')

            if data['participant_uuid'] == participant.participant_uuid:
                participant.authenticated = True
                return participant
            else:
                return None

        return None

    @staticmethod
    def get_participant_by_uuid(p_uuid, with_deleted: bool = False):
        participant = TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_uuid=p_uuid).first()

        if participant:
            return participant

        return None

    @staticmethod
    def get_participant_by_username(username, with_deleted: bool = False):
        return TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_username=username).first()

    @staticmethod
    def get_participant_by_email(email: str, with_deleted: bool = False):
        return TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_email=email).first()

    @staticmethod
    def get_participant_by_name(name, with_deleted: bool = False):
        return TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_name=name).first()

    @staticmethod
    def get_participant_by_id(part_id: int, with_deleted: bool = False):
        return TeraParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_participant=part_id).first()

    @staticmethod
    def is_participant_username_available(username: str) -> bool:
        # No username = always available
        if username is None or username == '':
            return True

        return TeraParticipant.query.filter_by(participant_username=username).first() is None

    @staticmethod
    def search_participant_by_name(name: str, other_filters: dict | None):
        if other_filters is None:
            other_filters = dict()
        search_query = (TeraParticipant.query.filter_by(**other_filters).
                        filter(TeraParticipant.participant_name.like('%' + name + '%')))
        return search_query.all()


    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraProject import TeraProject
            project1 = TeraProject.get_project_by_projectname('Default Project #1')
            project2 = TeraProject.get_project_by_projectname('Secret Project #1')

            participant1 = TeraParticipant()
            participant1.participant_name = 'Participant #1'
            participant1.participant_enabled = True
            participant1.participant_uuid = str(uuid.uuid4())
            participant1.participant_participant_group = \
                TeraParticipantGroup.get_participant_group_by_group_name('Default Participant Group A')
            participant1.participant_project = project1

            # participant1.create_token()
            participant1.participant_username = 'participant1'
            participant1.participant_password = TeraParticipant.encrypt_password('opentera')
            participant1.participant_login_enabled = True
            participant1.participant_token_enabled = True

            TeraParticipant.db().session.add(participant1)

            participant2 = TeraParticipant()
            participant2.participant_name = 'Participant #2'
            participant2.participant_enabled = False
            participant2.participant_uuid = str(uuid.uuid4())
            participant2.participant_participant_group = None
            participant2.participant_project = project1

            TeraParticipant.db().session.add(participant2)

            participant2 = TeraParticipant()
            participant2.participant_name = 'Participant #3'
            participant2.participant_enabled = True
            participant2.participant_token_enabled = True
            participant2.participant_uuid = str(uuid.uuid4())
            participant2.participant_participant_group = None
            participant2.participant_project = project1

            # participant2.create_token()
            TeraParticipant.db().session.add(participant2)

            participant2 = TeraParticipant()
            participant2.participant_name = 'Secret Participant'
            participant2.participant_enabled = True
            participant2.participant_token_enabled = True
            participant2.participant_uuid = str(uuid.uuid4())
            participant2.participant_participant_group = None
            participant2.participant_project = project2

            # participant2.create_token()
            TeraParticipant.db().session.add(participant2)

            TeraParticipant.db().session.commit()

            # Create token with added participants, since we need to have the id_participant field set
            participant1.create_token()
            participant2.create_token()
            TeraParticipant.db().session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        update_participant = TeraParticipant.get_participant_by_id(update_id)
        # Check if participant is an enabled project
        if 'participant_enabled' in values and values['participant_enabled'] and \
                not update_participant.participant_project.project_enabled:
            raise IntegrityError('Participant project disabled - no update allowed', update_id, 't_projects')

        # Check if username is available
        if 'participant_username' in values:
            if not TeraParticipant.is_participant_username_available(values['participant_username']):
                raise NameError('Participant username already in use.')

        # Prevent changes on UUID
        if 'participant_uuid' in values:
            del values['participant_uuid']

        # Remove the password field is present and if empty
        if 'participant_password' in values:
            if values['participant_password'] == '':
                del values['participant_password']
            else:
                # Forcing password to string
                values['participant_password'] = TeraParticipant.encrypt_password(str(values['participant_password']))

        # Check if we need to generate or delete tokens
        if 'participant_token_enabled' in values:
            if values['participant_token_enabled'] != update_participant.participant_token_enabled:
                if 'participant_enabled' in values:
                    participant_enabled = values['participant_enabled']
                else:
                    participant_enabled = update_participant.participant_enabled
                # Value changed
                if not values['participant_token_enabled'] or not participant_enabled:
                    values['participant_token'] = None  # Remove token
                else:
                    values['participant_token'] = update_participant.create_token()  # Generate new token
                    TeraParticipant.db().session.rollback()  # Don't save token here

        super().update(update_id, values)

    @classmethod
    def insert(cls, participant):
        # Encrypts password
        if participant.participant_password:
            # Forcing password to string
            participant.participant_password = TeraParticipant.encrypt_password(str(participant.participant_password))

        participant.participant_lastonline = None
        participant.participant_uuid = str(uuid.uuid4())
        participant.participant_token = None
        # participant.create_token()
        # Check if username is available
        if not TeraParticipant.is_participant_username_available(participant.participant_username):
            raise NameError('Participant username already in use.')
        super().insert(participant)

        # Token must be created after being inserted, since we need to have a valid ID participant into it
        if participant.participant_token_enabled and participant.participant_enabled:
            participant.create_token()

        # Check if participant project is enabled
        project = TeraProject.get_project_by_id(participant.id_project)
        if not project or not project.project_enabled:
            raise IntegrityError('Participant project disabled - no insert allowed', -1, 't_projects')
        TeraParticipant.db().session.commit()

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        # Safety check - can't delete participants with sessions
        if TeraSessionParticipants.get_session_count_for_participant(self.id_participant,
                                                                     with_deleted=with_deleted) > 0:
            return IntegrityError('Participant still has sessions', self.id_participant, 't_sessions_participants')

        if TeraSession.get_count(filters={'id_creator_participant': self.id_participant},
                                 with_deleted=with_deleted) > 0:
            return IntegrityError('Participant still has created sessions', self.id_participant, 't_sessions')

        if TeraAsset.get_count(filters={'id_participant': self.id_participant}, with_deleted=with_deleted) > 0:
            return IntegrityError('Participant still has created assets', self.id_participant, 't_assets')

        if TeraTest.get_count(filters={'id_participant': self.id_participant}, with_deleted=with_deleted) > 0:
            return IntegrityError('Participant still has created tests', self.id_participant, 't_tests')

        return None

    def hard_delete_before(self):
        # Delete sessions that we are part of since they will not be deleted otherwise
        for ses in self.participant_sessions:
            ses.hard_delete()

    def get_undelete_cascade_relations(self):
        return ['participant_service_config']



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.db.models.TeraSessionType import TeraSessionType


class TeraSessionTypeForm:

    @staticmethod
    def get_session_type_form(services: list):
        form = TeraForm("session_type")

        # Building lists
        categories = TeraSessionType.SessionCategoryEnum
        categories_list = []
        for category in categories:
            name = gettext(TeraSessionType.get_category_name(category))
            categories_list.append(TeraFormValue(value_id=category.value, value=name))

        services_list = []
        for service in services:
            services_list.append(TeraFormValue(value_id=service.id_service, value=service.service_name))

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_session_type", gettext("Session Type ID"), "hidden", True))
        section.add_item(TeraFormItem("session_type_name", gettext("Session Type Name"), "text", True))
        # section.add_item(TeraFormItem("session_type_prefix", gettext("Session Type Prefix"), "text", True,
        #                               item_options={'max_length': 10}))
        section.add_item(TeraFormItem("session_type_category", gettext("Category"), "array", item_required=True,
                                      item_values=categories_list))
        section.add_item(TeraFormItem("id_service", gettext("Service"), "array", item_required=False,
                                      item_values=services_list,
                                      item_condition= TeraFormItemCondition("session_type_category", "=",
                                                                            TeraSessionType.SessionCategoryEnum.SERVICE
                                                                            .value)
                                      )
                         )
        section.add_item(TeraFormItem("session_type_service_key", gettext("Session Type Service Key"), "hidden", False))
        section.add_item(TeraFormItem("session_type_online", gettext("Session Type Online"), "boolean", True))
        section.add_item(TeraFormItem("session_type_color", gettext("Session Type Display Color"), "color", True))
        section.add_item(TeraFormItem("session_type_config", gettext("Session Type Configuration"), "longtext", False))

        return form.to_dict()



from opentera.services.ServiceOpenTera import ServiceOpenTera
from opentera.modules.BaseModule import ModuleNames, create_module_message_topic_from_name, \
    create_module_event_topic_from_name
from opentera.db.models.TeraSession import TeraSessionStatus
from opentera.db.models.TeraSessionEvent import TeraSessionEvent

# Twisted
from twisted.internet import defer
from twisted.python import log
import opentera.messages.python as messages
import sys
import uuid
import json

from flask_babel import gettext

# Configuration
from ConfigManager import ConfigManager

from opentera.redis.RedisVars import RedisVars

# Modules
from requests import Response

# Messages
from google.protobuf.json_format import ParseError
from google.protobuf.message import DecodeError

"""
This class should be moved to opentera/service
"""


class BaseWebRTCService(ServiceOpenTera):
    def __init__(self, config_man: ConfigManager, service_info: dict):
        ServiceOpenTera.__init__(self, config_man, service_info)

        # This variable must be updated in derived classes
        self.webRTCModule = None

        # Active sessions
        self.sessions = dict()

    @defer.inlineCallbacks
    def register_to_events(self):
        # Need to register to events produced by UserManagerModule
        yield self.subscribe_pattern_with_callback(create_module_event_topic_from_name(
            ModuleNames.USER_MANAGER_MODULE_NAME), self.user_manager_event_received)

        super().register_to_events()

    def send_join_message(self, session_info, join_msg: str = gettext('Join me!'), target_users: list = None,
                          target_participants: list = None, target_devices: list = None):

        self.logger.log_info(self.config['name'],
                             'send_join_message',
                             session_info['session_uuid'],
                             target_users,
                             target_participants,
                             target_devices)

        users = session_info['session_users']
        participants = session_info['session_participants']
        devices = session_info['session_devices']
        parameters = session_info['session_parameters']

        join_message = messages.JoinSessionEvent()

        # Order is important here if multiple session creator
        if session_info['id_creator_service']:
            join_message.session_creator_name = session_info['session_creator_service']
        if session_info['id_creator_device']:
            join_message.session_creator_name = session_info['session_creator_device']
        if session_info['id_creator_participant']:
            join_message.session_creator_name = session_info['session_creator_participant']
        if session_info['id_creator_user']:
            join_message.session_creator_name = session_info['session_creator_user']
        join_message.session_uuid = session_info['session_uuid']
        for user_uuid in users:
            join_message.session_users.extend([user_uuid])
        for participant_uuid in participants:
            join_message.session_participants.extend([participant_uuid])
        for device_uuid in devices:
            join_message.session_devices.extend([device_uuid])
        join_message.join_msg = join_msg
        if not parameters:
            parameters = {}

        # Conversion to str
        if type(parameters) is dict:
            parameters = json.dumps(parameters)

        join_message.session_parameters = parameters

        join_message.service_uuid = self.service_uuid

        # Send invitations (as events) for users, participants and devices
        if target_users is None:
            target_users = users
        if target_devices is None:
            target_devices = devices
        if target_participants is None:
            target_participants = participants

        for user_uuid in target_users:
            join_message.session_url = session_info['session_url_users']
            self.send_event_message(join_message, 'websocket.user.'
                                    + user_uuid + '.events')
        for participant_uuid in target_participants:
            join_message.session_url = session_info['session_url_participants']
            self.send_event_message(join_message, 'websocket.participant.'
                                    + participant_uuid + '.events')
        for device_uuid in target_devices:
            join_message.session_url = session_info['session_url_devices']
            self.send_event_message(join_message, 'websocket.device.'
                                    + device_uuid + '.events')

        # Send event to UserManager to change "busy" status
        self.send_tera_message(join_message, 'service' + self.service_info['service_key'],
                               create_module_message_topic_from_name(ModuleNames.USER_MANAGER_MODULE_NAME))

    def user_manager_event_received(self, pattern, channel, message):
        # print('BaseWebRTCService - user_manager_event_received', pattern, channel, message)
        try:
            tera_event = messages.TeraEvent()
            if isinstance(message, str):
                ret = tera_event.ParseFromString(message.encode('utf-8'))
            elif isinstance(message, bytes):
                ret = tera_event.ParseFromString(message)

            user_event = messages.UserEvent()
            participant_event = messages.ParticipantEvent()
            device_event = messages.DeviceEvent()

            # Look for UserEvent, ParticipantEvent, DeviceEvent
            for any_msg in tera_event.events:
                if any_msg.Unpack(user_event):
                    self.handle_user_event(user_event)
                if any_msg.Unpack(participant_event):
                    self.handle_participant_event(participant_event)
                if any_msg.Unpack(device_event):
                    self.handle_device_event(device_event)

        except DecodeError as d:
            print('BaseWebRTCService - DecodeError ', pattern, channel, message, d)
        except ParseError as e:
            print('BaseWebRTCService - Failure in redisMessageReceived', e)

    def handle_user_event(self, event: messages.UserEvent):
        # print('BaseWebRTCService.handle_user_event', event)
        # Verify each session
        for id_session in self.sessions:
            session_info = self.sessions[id_session]

            # Verify if it contains the user_uuid
            if event.user_uuid in session_info['session_users']:
                # Verify the event type
                if event.type == messages.UserEvent.USER_CONNECTED:
                    # Resend invitation to newly connected user
                    # print('Resending invitation to ', event, session_info)

                    self.send_join_message(session_info=session_info, target_devices=[], target_participants=[],
                                           target_users=[event.user_uuid])

                elif event.type == messages.UserEvent.USER_DISCONNECTED:
                    # Terminate session if last user ?
                    if 'session_creator_user_uuid' in session_info:
                        if session_info['session_creator_user_uuid'] == event.user_uuid:
                            manage_session_args = {'id_session': id_session}
                            self.manage_stop_session(manage_session_args)
                            # End loop, sessions dict will be changed in manage_stop_session
                            break

    def handle_participant_event(self, event: messages.ParticipantEvent):
        # print('BaseWebRTCService.handle_participant_event', event)
        # Verify each session
        for id_session in self.sessions:
            session_info = self.sessions[id_session]

            # Verify if it contains the user_uuid
            if event.participant_uuid in session_info['session_participants']:
                # Verify the event type
                if event.type == messages.ParticipantEvent.PARTICIPANT_CONNECTED:
                    # Resend invitation to newly connected user
                    # print('Resending invitation to ', event, session_info)

                    self.send_join_message(session_info=session_info, target_devices=[],
                                           target_participants=[event.participant_uuid], target_users=[])

                elif event.type == messages.ParticipantEvent.PARTICIPANT_DISCONNECTED:
                    # End session if the participant was the creator
                    if 'session_creator_participant_uuid' in session_info:
                        if session_info['session_creator_participant_uuid'] == event.participant_uuid:
                            manage_session_args = {'id_session': id_session}
                            self.manage_stop_session(manage_session_args)
                            # End loop, sessions dict will be changed in manage_stop_session
                            break

    def handle_device_event(self, event: messages.DeviceEvent):
        # print('BaseWebRTCService.handle_device_event', event)
        # Verify each session
        for id_session in self.sessions:
            session_info = self.sessions[id_session]

            # Verify if it contains the user_uuid
            if event.device_uuid in session_info['session_devices']:
                # Verify the event type
                if event.type == messages.DeviceEvent.DEVICE_CONNECTED:
                    # Resend invitation to newly connected user
                    # print('Resending invitation to ', event, session_info)

                    self.send_join_message(session_info=session_info, target_devices=[event.device_uuid],
                                           target_participants=[], target_users=[])

                elif event.type == messages.DeviceEvent.DEVICE_DISCONNECTED:
                    # End session if the device was the creator
                    if 'session_creator_device_uuid' in session_info:
                        if session_info['session_creator_device_uuid'] == event.device_uuid:
                            manage_session_args = {'id_session': id_session}
                            self.manage_stop_session(manage_session_args)
                            # End loop, sessions dict will be changed in manage_stop_session
                            break

    def nodejs_webrtc_message_callback(self, pattern, channel, message):
        # print('WebRTCModule - nodejs_webrtc_message_callback', pattern, channel, message)
        parts = channel.split(".")
        if len(parts) == 2:
            session_key = parts[1]
            # print(session_key)
            if message == 'Ready!':
                # print('Ready!')
                self.handle_nodejs_session_ready(session_key)

    def get_session_info_from_key(self, session_key):
        for session_id in self.sessions:
            if self.sessions[session_id]['session_key'] == session_key:
                return self.sessions[session_id]
        return None

    def handle_nodejs_session_ready(self, session_key):
        # Should send invitations
        session_info = self.get_session_info_from_key(session_key)

        if session_info:
            self.send_join_message(session_info=session_info)

    def setup_rpc_interface(self):
        super().setup_rpc_interface()

        self.rpc_api['session_manage'] = {'args': ['str:json_info'],
                                          'returns': 'dict',
                                          'callback': self.session_manage}

        self.rpc_api['participant_info'] = {'args': ['str:participant_uuid'],
                                            'returns': 'dict',
                                            'callback': self.participant_info}

    def participant_info(self, participant_uuid):
        # Getting participant info from uuid
        params = {'participant_uuid': participant_uuid}
        response = self.get_from_opentera('/api/service/participants', params)
        if response.status_code == 200:
            return response.json()
        return {}

    def post_session_event(self, event_type: int, id_session: int, event_text: str = None) -> Response:
        from datetime import datetime
        api_req = {'session_event': {'id_session_event': 0,
                                     'id_session': id_session,
                                     'id_session_event_type': event_type,  # START session event
                                     'session_event_datetime': str(datetime.now()),
                                     'session_event_context': self.service_info['service_key']
                                     }
                   }

        if event_text:
            api_req['session_event']['session_event_text'] = event_text

        return self.post_to_opentera('/api/service/sessions/events', api_req)

    def session_manage(self, json_str):

        # - Service will create session if needed or reuse existing one
        # - Service will send invitations / updates to participants / users
        # - Service will return id_session and status code as a result to the RPC call and this will be the reply of
        #   this query
        import json
        try:

            session_manage = json.loads(json_str)

            if 'session_manage' in session_manage:
                session_manage_args = session_manage['session_manage']
                # Get all arguments
                # TODO VALIDATE ARGUMENTS

                # Get "common" arguments
                action = session_manage_args['action']

                if action == 'start':
                    return self.manage_start_session(session_manage_args)
                elif action == 'stop':
                    return self.manage_stop_session(session_manage_args)
                elif action == 'invite':
                    return self.manage_invite_to_session(session_manage_args)
                elif action == 'remove':
                    return self.manage_remove_from_session(session_manage_args)
                elif action == 'invite_reply':
                    return self.manage_invite_reply(session_manage_args)

        except json.JSONDecodeError as e:
            print('Error', e)
            return None

        return None

    def manage_start_session(self, session_manage_args: dict):
        # Get "useful" arguments
        id_service = session_manage_args['id_service']
        id_session_type = session_manage_args['id_session_type']
        id_session = session_manage_args['id_session']
        parameters = None
        if 'parameters' in session_manage_args:
            parameters = session_manage_args['parameters']

        # Get additional "start" arguments
        if 'session_participants' in session_manage_args:
            participants = session_manage_args['session_participants']
        else:
            participants = []
        if 'session_users' in session_manage_args:
            users = session_manage_args['session_users']
        else:
            users = []
        if 'session_devices' in session_manage_args:
            devices = session_manage_args['session_devices']
        else:
            devices = []

        # Call service API to create session
        api_response = None
        if id_session == 0:  # New session request
            api_req = {'id_session': 0,  # New session
                       'id_session_type': id_session_type,
                       'session_participants_uuids': participants,
                       'session_users_uuids': users,
                       'session_devices_uuids': devices,
                       'session_parameters': parameters
                       }
            if 'id_creator_user' in session_manage_args:
                api_req['id_creator_user'] = session_manage_args['id_creator_user']

            if 'id_creator_participant' in session_manage_args:
                api_req['id_creator_participant'] = session_manage_args['id_creator_participant']

            if 'id_creator_device' in session_manage_args:
                api_req['id_creator_device'] = session_manage_args['id_creator_device']

            if 'id_creator_service' in session_manage_args:
                api_req['id_creator_service'] = session_manage_args['id_creator_service']

            api_req = {'session': api_req}
            api_response = self.post_to_opentera('/api/service/sessions', api_req)
        else:
            api_response = self.get_from_opentera('/api/service/sessions', {'id_session': str(id_session),
                                                                            'with_events': True})

        if api_response.status_code == 200:

            session_info = api_response.json()

            if isinstance(session_info, list):
                session_info = session_info.pop()

            if 'session_events' not in session_info:
                session_info['session_events'] = []

            # Create start event in session events
            api_response = self.post_session_event(event_type=TeraSessionEvent.SessionEventTypes.SESSION_START.value
                                                   , id_session=session_info['id_session'])

            if api_response.status_code != 200:
                return {'status': 'error', 'error_text': gettext('Cannot create session event')}

            # Add event to list
            new_event = api_response.json()

            if isinstance(new_event, list):
                new_event = new_event.pop()
            session_info['session_events'].append(new_event)

            # Replace fields with uuids if presents
            session_info['session_participants'] = participants
            session_info['session_users'] = users
            session_info['session_devices'] = devices

            # Add session key
            session_info['session_key'] = str(uuid.uuid4())

            # Get session creator uuid
            creator_uuid = None
            if 'session_creator_user_uuid' in session_info:
                creator_uuid = session_info['session_creator_user_uuid']
            elif 'session_creator_participant_uuid' in session_info:
                creator_uuid = session_info['session_creator_participant_uuid']
            elif 'session_creator_device_uuid' in session_info:
                creator_uuid = session_info['session_creator_device_uuid']
            elif 'session_creator_service_uuid' in session_info:
                creator_uuid = session_info['session_creator_service_uuid']

            session_info['session_creator_uuid'] = creator_uuid

            # New WebRTC process with send events on this pattern
            self.subscribe_pattern_with_callback('webrtc.' + session_info['session_key'],
                                                 self.nodejs_webrtc_message_callback)

            # Start WebRTC process
            retval, process_info = self.webRTCModule.create_webrtc_session(session_info)

            if not retval or not process_info:
                self.unsubscribe_pattern_with_callback('webrtc.' + session_info['session_key'],
                                                       self.nodejs_webrtc_message_callback)
                return {'status': 'error', 'error_text': gettext('Cannot create process')}

            # Add URL to session_info
            session_info['session_url_users'] = process_info['url_users']
            session_info['session_url_participants'] = process_info['url_participants']
            session_info['session_url_devices'] = process_info['url_devices']

            # Keep session info for future use
            self.sessions[session_info['id_session']] = session_info

            # Return session information
            return {'status': 'started', 'session': session_info}

        else:
            return {'status': 'error', 'error_text': gettext('Cannot create session') + ': ' + api_response.text}

    def manage_stop_session(self, session_manage_args: dict):
        id_session = session_manage_args['id_session']
        # id_service = session_manage_args['id_service']
        # id_creator_user = session_manage_args['id_creator_user']

        if id_session in self.sessions:
            session_info = self.sessions[id_session]
            # Room name = key
            self.webRTCModule.stop_webrtc_session(session_info['session_key'])

            # Unsubscribe to messages from this process
            self.unsubscribe_pattern_with_callback('webrtc.' + session_info['session_key'],
                                                   self.nodejs_webrtc_message_callback)

            # Call service API for session changes...

            # Create session stop event
            api_response = self.post_session_event(event_type=TeraSessionEvent.SessionEventTypes.SESSION_STOP.value,
                                                   id_session=session_info['id_session'])

            if api_response.status_code != 200:
                return {'status': 'error', 'error_text': gettext('Cannot create STOP session event')}

            # Compute session duration from last start event
            from datetime import datetime
            duration = 0

            for session_event in reversed(session_info['session_events']):
                if session_event['id_session_event_type'] == 3:  # START event
                    time_diff = datetime.now() - datetime.fromisoformat(session_event['session_event_datetime']).\
                        replace(tzinfo=None)
                    duration = int(time_diff.total_seconds())
                    break

            # Default duration
            # if duration == 0:
            #     time_diff = datetime.now() - datetime.fromisoformat(session_info['session_start_datetime']).\
            #         replace(tzinfo=None)
            #     duration = int(abs(time_diff.total_seconds()))  # abs in case of sessions set in the future

            # Add current session duration to the total
            duration += session_info['session_duration']

            # Call service API to update session
            api_req = {'session': {'id_session': id_session,
                                   'session_status': TeraSessionStatus.STATUS_COMPLETED.value,
                                   'session_duration': duration}}

            api_response = self.post_to_opentera('/api/service/sessions', api_req)

            # Send events
            stop_session_event = messages.StopSessionEvent()
            stop_session_event.session_uuid = session_info['session_uuid']
            stop_session_event.service_uuid = self.service_uuid

            for user_uuid in session_info['session_users']:
                self.send_event_message(stop_session_event, 'websocket.user.' + user_uuid + '.events')

            for participant_uuid in session_info['session_participants']:
                self.send_event_message(stop_session_event, 'websocket.participant.' + participant_uuid + '.events')

            for device_uuid in session_info['session_devices']:
                self.send_event_message(stop_session_event, 'websocket.device.' + device_uuid + '.events')

            # Send event to UserManager to change "busy" status
            self.send_tera_message(stop_session_event, 'service' + self.service_info['service_key'],
                                   create_module_message_topic_from_name(ModuleNames.USER_MANAGER_MODULE_NAME))

            # Remove session from list
            del self.sessions[id_session]

            # Return response
            if api_response.status_code == 200:
                session_info = api_response.json()
                if isinstance(session_info, list):
                    session_info = session_info.pop()
                return {'status': 'stopped', 'session': session_info}

            return {'status': 'error', 'error_text': gettext('Error stopping session - check server logs')}

        return {'status': 'error', 'error_text': gettext('No matching session to stop')}

    def manage_invite_to_session(self, session_manage_args: dict):
        id_session = session_manage_args['id_session']
        # id_service = session_manage_args['id_service']
        # id_creator_user = session_manage_args['id_creator_user']

        if id_session in self.sessions:
            session_info = self.sessions[id_session]

            new_session_users = []
            new_session_devices = []
            new_session_participants = []

            if 'session_users' in session_manage_args:
                new_session_users = session_manage_args['session_users']
                session_info['session_users'].extend(new_session_users)

                for session_user in new_session_users:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN.value, id_session=id_session,
                        event_text=gettext('User') + ': ' + session_user)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating user invited session event')}

            if 'session_participants' in session_manage_args:
                new_session_participants = session_manage_args['session_participants']
                session_info['session_participants'].extend(new_session_participants)
                for session_participant in new_session_participants:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN.value,
                        id_session=id_session,
                        event_text=gettext('Participant') + ': ' + session_participant)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating participant invited '
                                                                         'session event')}

            if 'session_devices' in session_manage_args:
                new_session_devices = session_manage_args['session_devices']
                session_info['session_devices'].extend(new_session_devices)
                for session_device in new_session_devices:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN.value,
                        id_session=id_session,
                        event_text=gettext('Device') + ': ' + session_device)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating device invited '
                                                                         'session event')}

            # Update session with new invitees
            api_req = {'session': {'id_session': id_session,  # New session
                                   'session_participants_uuids': session_info['session_participants'],
                                   'session_users_uuids': session_info['session_users'],
                                   'session_devices_uuids': session_info['session_devices'],
                                   }
                       }
            api_response = self.post_to_opentera('/api/service/sessions', api_req)
            if api_response.status_code == 200:
                # Resend join session message to all invitees to let them update their list if needed
                self.send_join_message(session_info=session_info)
                return {'status': 'invited', 'session': session_info}

            return {'status': 'error', 'error_text': gettext('Error updating session')}

    def manage_remove_from_session(self, session_manage_args: dict):
        id_session = session_manage_args['id_session']
        # id_service = session_manage_args['id_service']
        # id_creator_user = session_manage_args['id_creator_user']

        if id_session in self.sessions:
            session_info = self.sessions[id_session]

            removed_session_users = []
            removed_session_devices = []
            removed_session_participants = []

            session_users = session_info['session_users']
            session_devices = session_info['session_devices']
            session_participants = session_info['session_participants']

            if 'session_users' in session_manage_args:
                removed_session_users = session_manage_args['session_users']
                session_info['session_users'] = [item for item in session_info['session_users']
                                                 if item not in removed_session_users]

                for session_user in removed_session_users:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_LEAVE.value,
                        id_session=id_session,
                        event_text=gettext('User') + ': ' + session_user)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating user left session event')}

            if 'session_participants' in session_manage_args:
                removed_session_participants = session_manage_args['session_participants']
                session_info['session_participants'] = [item for item in session_info['session_participants']
                                                        if item not in removed_session_participants]
                for session_participant in removed_session_participants:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_LEAVE.value,
                        id_session=id_session,
                        event_text=gettext('Participant') + ': ' + session_participant)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating participant left '
                                                                         'session event')}

            if 'session_devices' in session_manage_args:
                removed_session_devices = session_manage_args['session_devices']
                session_info['session_devices'] = [item for item in session_info['session_devices']
                                                   if item not in removed_session_devices]
                for session_device in removed_session_devices:
                    # Get names for log
                    # TODO
                    #     api_response = self.get_from_opentera('/api/service')
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_LEAVE.value,
                        id_session=id_session,
                        event_text=gettext('Device') + ': ' + session_device)
                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Error creating device left '
                                                                         'session event')}

            # Create and send leave session event message
            leave_message = messages.LeaveSessionEvent()
            leave_message.session_uuid = session_info['session_uuid']
            leave_message.service_uuid = self.service_uuid
            for user_uuid in removed_session_users:
                leave_message.leaving_users.extend([user_uuid])
            for participant_uuid in removed_session_participants:
                leave_message.leaving_participants.extend([participant_uuid])
            for device_uuid in removed_session_devices:
                leave_message.leaving_devices.extend([device_uuid])

            # Broadcast to all
            for user_uuid in session_users:
                self.send_event_message(leave_message, 'websocket.user.' + user_uuid + '.events')
            for participant_uuid in session_participants:
                self.send_event_message(leave_message, 'websocket.participant.' + participant_uuid + '.events')
            for device_uuid in session_devices:
                self.send_event_message(leave_message, 'websocket.device.' + device_uuid + '.events')

            # Send event to UserManager to change "busy" status
            self.send_tera_message(leave_message, 'service' + self.service_info['service_key'],
                                   create_module_message_topic_from_name(ModuleNames.USER_MANAGER_MODULE_NAME))

            # Don't update session with current list of users - participants... We need to keep a trace that they
            # were part of that session at some point!

            return {'status': 'removed', 'session': session_info}

    def manage_invite_reply(self, session_manage_args: dict):
        id_session = session_manage_args['id_session']
        parameters = session_manage_args['parameters']
        reply_code = parameters['reply_code']

        if id_session in self.sessions:
            session_info = self.sessions[id_session]

            join_session_reply = messages.JoinSessionReplyEvent()
            join_session_reply.session_uuid = session_info['session_uuid']
            join_session_reply.join_reply = parameters['reply_code']
            if 'reply_msg' in parameters:
                join_session_reply.reply_msg = parameters['reply_msg']
            if 'user_uuid' in parameters:
                join_session_reply.user_uuid = parameters['user_uuid']
                if parameters['reply_code'] != messages.JoinSessionReplyEvent.REPLY_ACCEPTED:
                    session_info['session_users'] = [item for item in session_info['session_users']
                                                     if item != parameters['user_uuid']]
                    # Create session join refused event
                    # TODO: Get user name instead of user uuid
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN_REFUSED.value,
                        id_session=session_info['id_session'],
                        event_text=gettext('User') + ' ' + parameters['user_uuid'])

                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Cannot create refused session event')}

            if 'participant_uuid' in parameters:
                join_session_reply.participant_uuid = parameters['participant_uuid']
                if parameters['reply_code'] != messages.JoinSessionReplyEvent.REPLY_ACCEPTED:
                    session_info['session_participants'] = [item for item in session_info['session_participants']
                                                            if item != parameters['participant_uuid']]
                    # Create session join refused event
                    # TODO: Get  participant name instead of uuid
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN_REFUSED.value,
                        id_session=session_info['id_session'],
                        event_text=gettext('Participant') + ' ' + parameters['participant_uuid'])

                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Cannot create refused session event')}

            if 'device_uuid' in parameters:
                join_session_reply.device_uuid = parameters['device_uuid']
                if parameters['reply_code'] != messages.JoinSessionReplyEvent.REPLY_ACCEPTED:
                    session_info['session_devices'] = [item for item in session_info['session_devices']
                                                       if item != parameters['device_uuid']]
                    # Create session join refused event
                    # TODO: Get device name instead of uuid
                    api_response = self.post_session_event(
                        event_type=TeraSessionEvent.SessionEventTypes.SESSION_JOIN_REFUSED.value,
                        id_session=session_info['id_session'],
                        event_text=gettext('Device') + ' ' + parameters['device_uuid'])

                    if api_response.status_code != 200:
                        return {'status': 'error', 'error_text': gettext('Cannot create refused session event')}

            # Send reply message to everyone
            session_users = session_info['session_users']
            session_devices = session_info['session_devices']
            session_participants = session_info['session_participants']
            for user_uuid in session_users:
                self.send_event_message(join_session_reply, 'websocket.user.' + user_uuid + '.events')
            for participant_uuid in session_participants:
                self.send_event_message(join_session_reply, 'websocket.participant.' + participant_uuid + '.events')
            for device_uuid in session_devices:
                self.send_event_message(join_session_reply, 'websocket.device.' + device_uuid + '.events')

            # Send event to UserManager to change "busy" status
            self.send_tera_message(join_session_reply, 'service' + self.service_info['service_key'],
                                   create_module_message_topic_from_name(ModuleNames.USER_MANAGER_MODULE_NAME))

            return {'status': 'OK', 'session': session_info}





from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.models.TeraTest import TeraTest
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError
import uuid


class TeraTestType(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_tests_types'
    id_test_type = Column(Integer, Sequence('id_test_type_sequence'), primary_key=True, autoincrement=True)
    id_service = Column(Integer, ForeignKey("t_services.id_service", ondelete='cascade'), nullable=False)
    test_type_uuid = Column(String(36), nullable=False, unique=True)
    test_type_name = Column(String, nullable=False, unique=False)
    test_type_description = Column(String, nullable=True)
    test_type_key = Column(String, nullable=True, unique=True)
    # Test type service provides TeraForm style json format?
    test_type_has_json_format = Column(Boolean, nullable=False, default=False)
    # Test type service provides a generated HTML form?
    test_type_has_web_format = Column(Boolean, nullable=False, default=False)
    # Test type service provides a web based editor?
    test_type_has_web_editor = Column(Boolean, nullable=False, default=False)

    test_type_test_type_projects = relationship("TeraTestTypeProject", viewonly=True)
    test_type_test_type_sites = relationship("TeraTestTypeSite", viewonly=True)

    test_type_service = relationship("TeraService")
    test_type_projects = relationship("TeraProject", secondary="t_tests_types_projects")
    test_type_sites = relationship("TeraSite", secondary="t_tests_types_sites")
    test_type_tests = relationship("TeraTest", cascade='delete')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['test_type_service', 'test_type_sites', 'test_type_projects',
                              'test_type_test_type_projects', 'test_type_test_type_sites', 'test_type_tests'])

        if minimal:
            ignore_fields.extend(['test_type_description'])
        rval = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            # Also includes service key and uuid
            rval['test_type_service_key'] = self.test_type_service.service_key
            rval['test_type_service_uuid'] = self.test_type_service.service_uuid
        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService

            test = TeraTestType()
            test.test_type_name = 'Pre-session evaluation'
            test.test_type_description = 'Evaluation shown before a session'
            test.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            test.test_type_uuid = str(uuid.uuid4())
            test.test_type_key = 'PRE'
            test.test_type_has_json_format = True
            TeraTestType.db().session.add(test)

            test = TeraTestType()
            test.test_type_name = 'Post-session evaluation'
            test.test_type_description = 'Evaluation shown after a session'
            test.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
            test.test_type_uuid = str(uuid.uuid4())
            test.test_type_has_json_format = True
            test.test_type_has_web_format = True
            TeraTestType.db().session.add(test)

            test = TeraTestType()
            test.test_type_name = 'General survey'
            test.test_type_description = 'General satisfaction survey'
            test.id_service = TeraService.get_service_by_key('FileTransferService').id_service
            test.test_type_uuid = str(uuid.uuid4())
            test.test_type_has_web_format = True
            test.test_type_has_web_editor = True
            TeraTestType.db().session.add(test)

            TeraTestType.db().session.commit()

    @staticmethod
    def get_test_type_by_id(test_type_id: int, with_deleted: bool = False):
        return TeraTestType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_test_type=test_type_id).first()

    @staticmethod
    def get_test_type_by_key(tt_key: int, with_deleted: bool = False):
        return TeraTestType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(test_type_key=tt_key).first()

    @staticmethod
    def get_test_type_by_uuid(tt_uuid: int, with_deleted: bool = False):
        return TeraTestType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(test_type_uuid=tt_uuid).first()

    @staticmethod
    def get_test_types_for_service(id_service: int, with_deleted: bool = False):
        return TeraTestType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=id_service).all()

    @staticmethod
    def get_access_token(test_type_uuids: list, token_key: str, requester_uuid: str, can_edit: bool, expiration=3600):
        import time
        import jwt

        # Creating token
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'test_type_uuids': test_type_uuids,
            'can_edit': can_edit,
            'requester_uuid': requester_uuid
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    def get_service_urls(self, server_url: str, server_port: int) -> dict:
        urls = {'test_type_json_url': None,
                'test_type_web_url': None,
                'test_type_web_editor_url': None}

        if not self.test_type_service.service_enabled:
            return urls  # Service disabled = no Urls!

        service_endpoint = self.test_type_service.service_clientendpoint
        base_url = 'https://' + server_url + ':' + str(server_port) + service_endpoint

        if self.test_type_has_json_format:
            urls['test_type_json_url'] = base_url + '/api/testtypes/form'

        if self.test_type_has_web_format:
            urls['test_type_web_url'] = base_url + '/api/testtypes/web'

        if self.test_type_has_web_editor:
            urls['test_type_web_editor_url'] = base_url + '/api/testtypes/web/edit'

        return urls

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        if TeraTest.get_count(filters={'id_test_type': self.id_test_type}, with_deleted=with_deleted) > 0:
            return IntegrityError('Test Type still has associated tests', self.id_test_type, 't_tests')
        return None

    @classmethod
    def insert(cls, test_type):
        # Generate UUID
        test_type.test_type_uuid = str(uuid.uuid4())

        super().insert(test_type)



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, Integer, String, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraUserGroup(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_users_groups'
    id_user_group = Column(Integer, Sequence('id_usergroup_sequence'), primary_key=True, autoincrement=True)
    user_group_name = Column(String, nullable=False, unique=False)

    user_group_services_roles = relationship("TeraServiceRole", secondary="t_services_access",
                                             back_populates="service_role_user_groups",
                                             passive_deletes=True)

    user_group_users = relationship("TeraUser", secondary="t_users_users_groups",
                                    back_populates="user_user_groups",
                                    passive_deletes=True)

    def __str__(self):
        return '<TeraUserGroup ' + str(self.user_group_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        if minimal:
            ignore_fields.extend([])
        return super().to_json(ignore_fields=ignore_fields)

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_user_group': self.id_user_group}

    def get_projects_roles(self,  service_id: int | None = None, no_inheritance: bool = False) -> dict:
        projects_roles = {}

        # Projects
        for service_role in self.user_group_services_roles:
            if service_id and service_role.id_service != service_id:
                # Need to limit to a specific service and this is not one for that
                continue
            if service_role.id_project:
                projects_roles[service_role.service_role_project] = \
                    {'project_role': service_role.service_role_name, 'inherited': False}

        # Sites - if we are admin in a site, we are automatically admin in all its project
        if not no_inheritance:
            for service_role in self.user_group_services_roles:
                if service_id and service_role.id_service != service_id:
                    # Need to limit to a specific service and this is not one for that
                    continue
                if service_role.id_site:
                    if service_role.service_role_name == 'admin':
                        for project in service_role.service_role_site.site_projects:
                            projects_roles[project] = {'project_role': 'admin', 'inherited': True}

        return projects_roles

    def get_sites_roles(self, service_id: int | None = None) -> dict:
        sites_roles = {}
        # Sites
        for service_role in self.user_group_services_roles:
            if service_id and service_role.id_service != service_id:
                # Need to limit to a specific service and this is not one for that
                continue
            if service_role.id_site:
                sites_roles[service_role.service_role_site] = \
                    {'site_role': service_role.service_role_name, 'inherited': False}

        # Projects - each project's site also provides a "user" access for that site
        for service_role in self.user_group_services_roles:
            if service_id and service_role.id_service != service_id:
                # Need to limit to a specific service and this is not one for that
                continue
            if service_role.id_project:
                project_site = service_role.service_role_project.project_site
                if project_site not in sites_roles:
                    sites_roles[project_site] = {'site_role': 'user', 'inherited': True}

        return sites_roles

    def get_global_roles(self, service_id: int) -> list:
        global_roles = []

        for service_role in self.user_group_services_roles:
            if not service_role.id_site and not service_role.id_project and service_role.id_service == service_id:
                global_roles.append(service_role.service_role_name)
        return global_roles

    @staticmethod
    def get_user_group_by_group_name(name: str, with_deleted: bool = False):
        return TeraUserGroup.query.execution_options(include_deleted=with_deleted)\
            .filter_by(user_group_name=name).first()

    @staticmethod
    def get_user_group_by_id(group_id: int, with_deleted: bool = False):
        return TeraUserGroup.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_user_group=group_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraProject import TeraProject
            from opentera.db.models.TeraServiceAccess import TeraServiceAccess
            from opentera.db.models.TeraServiceRole import TeraServiceRole
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraSite import TeraSite

            opentera_service_id = TeraService.get_openteraserver_service().id_service

            ugroup = TeraUserGroup()
            ugroup.user_group_name = "Users - Projects 1 & 2"
            TeraUserGroup.db().session.add(ugroup)

            ugroup = TeraUserGroup()
            ugroup.user_group_name = "Admins - Project 1"
            TeraUserGroup.db().session.add(ugroup)

            ugroup = TeraUserGroup()
            ugroup.user_group_name = "Admins - Default Site"
            TeraUserGroup.db().session.add(ugroup)

            ugroup = TeraUserGroup()
            ugroup.user_group_name = "Users - Project 1"
            TeraUserGroup.db().session.add(ugroup)

            ugroup = TeraUserGroup()
            ugroup.user_group_name = "No access!"
            TeraUserGroup.db().session.add(ugroup)

            TeraUserGroup.db().session.commit()

            id_user_group = TeraUserGroup.get_user_group_by_group_name('Users - Projects 1 & 2').id_user_group
            access = TeraServiceAccess()
            access.id_user_group = id_user_group
            id_project = TeraProject.get_project_by_projectname('Default Project #1').id_project
            user_role = TeraServiceRole.get_service_role_by_name(service_id=opentera_service_id, project_id=id_project,
                                                                 rolename='user')
            access.id_service_role = user_role.id_service_role
            TeraUserGroup.db().session.add(access)

            access = TeraServiceAccess()
            access.id_user_group = id_user_group
            id_project = TeraProject.get_project_by_projectname('Default Project #2').id_project
            user_role = TeraServiceRole.get_service_role_by_name(service_id=opentera_service_id, project_id=id_project,
                                                                 rolename='user')
            access.id_service_role = user_role.id_service_role
            TeraUserGroup.db().session.add(access)

            admin_access = TeraServiceAccess()
            admin_role = TeraServiceRole.get_service_role_by_name(service_id=opentera_service_id,
                                                                  site_id=TeraSite.get_site_by_sitename('Default Site')
                                                                  .id_site,
                                                                  rolename='admin')
            admin_access.id_service_role = admin_role.id_service_role
            admin_access.id_user_group = \
                TeraUserGroup.get_user_group_by_group_name('Admins - Default Site').id_user_group
            TeraUserGroup.db().session.add(admin_access)

            access = TeraServiceAccess()
            access.id_user_group = TeraUserGroup.get_user_group_by_group_name('Admins - Project 1').id_user_group
            id_project = TeraProject.get_project_by_projectname('Default Project #1').id_project
            admin_role = TeraServiceRole.get_service_role_by_name(service_id=opentera_service_id, project_id=id_project,
                                                                  rolename='admin')
            access.id_service_role = admin_role.id_service_role
            TeraUserGroup.db().session.add(access)

            access = TeraServiceAccess()
            access.id_user_group = TeraUserGroup.get_user_group_by_group_name('Users - Project 1').id_user_group
            user_role = TeraServiceRole.get_service_role_by_name(service_id=opentera_service_id, project_id=id_project,
                                                                 rolename='user')
            access.id_service_role = user_role.id_service_role
            TeraUserGroup.db().session.add(access)
            TeraUserGroup.db().session.commit()

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        if len(self.user_group_users) > 0:
            return IntegrityError('User group still has associated users', self.id_user_group, 't_users')
        return None
